"""Functions for creating spatial analysis plots for MPX data.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from pixelator.mpx.analysis.colocalization import get_differential_colocalization
from pixelator.mpx.analysis.polarization import get_differential_polarity

logger = logging.getLogger(__name__)


def _pivot_colocalization_data(
    colocalization_data: pd.DataFrame,
    value_col: str = "pearson",
    markers: Union[list, None] = None,
):
    colocalization_data_pivot = pd.pivot_table(
        colocalization_data,
        index="marker_1",
        columns="marker_2",
        values=value_col,
        fill_value=0,
    )

    if markers is not None:
        colocalization_data_pivot = colocalization_data_pivot.loc[markers, markers]

    for m in colocalization_data_pivot.index:
        colocalization_data_pivot.loc[m, m] = 0  # remove autocorrelations

    colocalization_data_pivot = colocalization_data_pivot.fillna(0)

    return colocalization_data_pivot


def _make_colocalization_symmetric(
    colocalization_data: pd.DataFrame,
    value_col: str = "pearson",
):
    colocalization_data = pd.DataFrame(
        np.concatenate(
            [
                colocalization_data[["marker_1", "marker_2", value_col]].to_numpy(),
                colocalization_data[["marker_2", "marker_1", value_col]].to_numpy(),
            ],
        ),
        columns=["marker_1", "marker_2", value_col],
    )
    colocalization_data = (
        colocalization_data.groupby(["marker_1", "marker_2"])[value_col]
        .apply(lambda x: np.mean(x))
        .reset_index()
    )

    return colocalization_data


def plot_colocalization_heatmap(
    colocalization_data: pd.DataFrame,
    markers: Union[list, None] = None,
    cmap: str = "vlag",
    value_column: str = "pearson_z",
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot a colocalization heatmap based on the provided colocalization data.

    The colocalization_data DataFrame should contain the columns "marker_1",
    "marker_2", "pearson", "pearson_z".
    Example usage: plot_colocalization_heatmap(pxl.colocalization).

    :param colocalization_data: The colocalization data to plot. The
    colocalization data frame that can be found in a pixel variable "pxl"
    through pxl.colocalization. The data frame should contain the columns
    "marker_1", "marker_2", "pearson", "pearson_z", and "component".
    :param markers: The markers to include in the heatmap. Defaults to None.
    :param cmap: The colormap to use for the heatmap. Defaults to "vlag".
    :param value_column: What colocalization metric to use. Defaults to "pearson_z".

    :return: The figure and axes objects of the plot.
    :rtype: Tuple[plt.Figure, plt.Axes]

    """
    colocalization_data = _make_colocalization_symmetric(
        colocalization_data, value_column
    )

    colocalization_data_pivot = _pivot_colocalization_data(
        colocalization_data, value_column, markers=markers
    )
    sns.clustermap(
        colocalization_data_pivot,
        yticklabels=True,
        xticklabels=True,
        linewidths=0.1,
        method="complete",
        cmap=cmap,
    )

    return plt.gcf(), plt.gca()


def plot_colocalization_diff_heatmap(
    colocalization_data: pd.DataFrame,
    reference: str,
    targets: str | list[str] | None = None,
    contrast_column: str = "sample",
    top_marker_log_p: float | None = None,
    min_log_p: float = 3.0,
    cmap: str = "vlag",
    value_column: str = "pearson_z",
) -> Tuple[dict, dict]:
    """Plot the differential colocalization between reference and target components.

    Example usage: plot_colocalization_diff_heatmap(pxl.colocalization,
    reference:"control", contrast_column="sample").

    :param colocalization_data: The colocalization data frame that can be found
    in a pixel variable "pxl" through pxl.colocalization. The data frame should
    contain the columns "marker_1", "marker_2", "pearson", "pearson_z", and the
    contrast_column.
    :param reference: The label for reference components in the contrast_column.
    :param targets: label or list of labels for target components in the
    contrast_column. When not specified, all labels in the contrast_column
    except the reference label are used as targets.
    :param contrast_column: The column to use for the contrast. Defaults to "sample".
    :param top_marker_log_p: When set to a value, only markers that differentially
    colocalize with at least one other marker with a log10 p-score higher than
    top_marker_log_p are inclueded in the plot.
    :param min_log_p: The minimum log10 p-value. Pairs with lower log10 p-value
    are assigned 0.
    :param cmap: The colormap to use for the heatmap. Defaults to "vlag".
    :param value_column: What colocalization metric to use. Defaults to "pearson_z".

    :return: Two dicts mapping target names respectively to figures and axes of
    the generated plots.
    """
    if isinstance(targets, str):
        targets = [targets]
    elif targets is None:
        targets = colocalization_data[contrast_column].unique()
        targets = list(set(targets) - {reference})

    differential_colocalization = get_differential_colocalization(
        colocalization_data,
        reference=reference,
        targets=targets,
        contrast_column=contrast_column,
        value_column=value_column,
    )
    figs = {}
    axes = {}
    for target in sorted(targets):
        target_diff = differential_colocalization.loc[
            differential_colocalization["target"] == target, :
        ]

        target_diff_p_adj = -np.log10(
            target_diff.set_index(["marker_1", "marker_2"])["p_adj"].unstack()
        )
        target_diff_med_diff = target_diff.set_index(["marker_1", "marker_2"])[
            "median_difference"
        ].unstack(fill_value=0)
        target_diff_med_diff[target_diff_p_adj < min_log_p] = 0
        if top_marker_log_p is not None:
            target_diff_med_diff = target_diff_med_diff.loc[
                target_diff_p_adj.max(axis=1) > top_marker_log_p,
                target_diff_p_adj.max(axis=0) > top_marker_log_p,
            ]
            if target_diff_med_diff.shape[0] == 0:
                logger.warning(
                    "No marker pairs with log10 p-value higher than "
                    f"{top_marker_log_p} found for target {target}."
                )
                continue
        target_diff_med_diff = target_diff_med_diff.add(
            target_diff_med_diff.T, fill_value=0
        ).fillna(0)
        max_value = np.max(np.abs(target_diff_med_diff.to_numpy().flatten()))
        g = sns.clustermap(
            target_diff_med_diff,
            yticklabels=True,
            xticklabels=True,
            method="complete",
            linewidths=0.1,
            vmin=-max_value,
            vmax=max_value,
            cmap=cmap,
        )
        g.figure.suptitle(
            f"Differential colocalization between {target} and {reference}"
        )
        plt.tight_layout()
        figs[target] = plt.gcf()
        axes[target] = plt.gca()

    return figs, axes


def _add_top_marker_labels(
    plot_data,
    ax,
    n_top_pairs: int = 5,
    min_log_p: float = 5.0,
):
    plot_data = plot_data.sort_values("median_difference")
    plot_data = plot_data.loc[-np.log10(plot_data["p_adj"]) > min_log_p, :]

    # Labels for marker pair withs highest negative differential colocalization scores
    for _, row in plot_data.head(n_top_pairs).iterrows():
        if "marker_1" not in row.index:
            if "marker" in row.index:
                name = row["marker"]
            else:
                name = row.name
        else:
            name = row["marker_1"] + "/" + row["marker_2"]

        x, y = row[["median_difference", "p_adj"]]
        y = -np.log10(y)
        if x > 0:
            continue
        ax.text(
            x,
            y,
            name,
            horizontalalignment="left",
            fontsize="xx-small",
        )

    # Labels for marker pair with highest positive differential colocalization scores
    for _, row in plot_data.tail(n_top_pairs).iterrows():
        if "marker_1" not in row.index:
            if "marker" in row.index:
                name = row["marker"]
            else:
                name = row.name
        else:
            name = row["marker_1"] + "/" + row["marker_2"]
        x, y = row[["median_difference", "p_adj"]]
        y = -np.log10(y)
        if x < 0:
            continue
        ax.text(
            x,
            y,
            name,
            horizontalalignment="left",
            fontsize="xx-small",
        )


def _add_target_mean_colocalizations(
    differential_colocalization,
    target_coloc,
    value_col,
):
    differential_colocalization = differential_colocalization.fillna(0).reset_index()

    target_values = (
        target_coloc.groupby(["marker_1", "marker_2"])[value_col].mean().reset_index()
    )
    differential_colocalization = pd.merge(
        target_values,
        differential_colocalization,
        on=["marker_1", "marker_2"],
        how="inner",
    )

    return differential_colocalization


def plot_colocalization_diff_volcano(
    colocalization_data: pd.DataFrame,
    reference: str,
    targets: str | list[str] | None = None,
    contrast_column: str = "sample",
    cmap: str = "vlag",
    value_column="pearson_z",
    n_top_pairs: int = 5,
    min_log_p: float = 5.0,
) -> Tuple[plt.Figure, plt.Axes]:
    """Generate the volcano plot of differential colocalization between reference and target(s) components.

    Example usage: `plot_colocalization_diff_volcano(pxl.colocalization, target:"stimulated", reference:"control", contrast_column="sample")`.

    :param colocalization_data: The colocalization data frame that can be found in a pixel variable
                                "pxl" through pxl.colocalization. The data frame should contain the
                                columns "marker_1", "marker_2", "pearson", "pearson_z", and the contrast_column.
    :param target: The label for target components in the contrast_column.
    :param reference: The label for reference components in the contrast_column.
    :param contrast_column: The column to use for the contrast. Defaults to "sample".
    :param cmap: The colormap to use for the heatmap. Defaults to "vlag".
    :param value_column: What colocalization metric to use. Defaults to "pearson_z".
    :param n_top_pairs: Number of high value marker-pairs to label from positive and negative sides.
    :param min_log_p: marker-pairs only receive a label if -log10 of their p-value is higher than
                      this parameter.

    :return: The figure and axes objects of the plot.
    """
    if isinstance(targets, str):
        targets = [targets]
    elif targets is None:
        targets = colocalization_data[contrast_column].unique()
        targets = list(set(targets) - {reference})

    if len(targets) > 5:
        raise ValueError(
            "Only up to 5 targets can be visualized. "
            "Number of requested targets is {len(targets)}."
            "Requested targets are: {targets}."
        )

    differential_colocalization = get_differential_colocalization(
        colocalization_data,
        targets=targets,
        reference=reference,
        contrast_column=contrast_column,
        value_column=value_column,
    )

    fig, axes = plt.subplots(1, len(targets))
    for i, target in enumerate(sorted(targets)):
        ax = axes[i] if len(targets) > 1 else axes
        target_differential_colocalization = differential_colocalization.loc[
            differential_colocalization["target"] == target, :
        ]
        target_differential_colocalization = _add_target_mean_colocalizations(
            target_differential_colocalization,
            target_coloc=colocalization_data.loc[
                colocalization_data[contrast_column] == target, :
            ],
            value_col=value_column,
        )
        p = ax.scatter(
            x=target_differential_colocalization["median_difference"],
            y=-np.log10(target_differential_colocalization["p_adj"]),
            c=target_differential_colocalization[value_column],
            s=20,
            marker="o",
            cmap=cmap,
        )

        ax.set(
            xlabel="Median difference",
            ylabel=r"$-\log_{10}$(adj. p-value)",
            title=f"Differential colocalization\nbetween {target}\nand {reference}",
        )
        ax.title.set_y(1.05)
        fig.colorbar(p, label="Mean target colocalization score", cmap=cmap)
        _add_top_marker_labels(
            target_differential_colocalization,
            n_top_pairs=n_top_pairs,
            min_log_p=min_log_p,
            ax=ax,
        )
    fig.subplots_adjust(top=0.8)
    fig.set_size_inches(6 * len(targets), 5)
    return fig, axes


def plot_polarity_diff_volcano(
    polarity_data: pd.DataFrame,
    reference: str,
    targets: str | list[str] | None = None,
    contrast_column: str = "sample",
    cmap: str = "vlag",
    value_column="morans_z",
    n_top_pairs: int = 5,
    min_log_p: float = 5.0,
) -> Tuple[plt.Figure, plt.Axes]:
    """Generate the volcano plot of differential polarity between reference and target(s) components.

    Example usage: `plot_polarity_diff_volcano(
                                                pxl.polariazation,target:"stimulated",
                                                reference:"control",
                                                contrast_column="sample"
                                                )`.

    :param polarity_data: The polarity data frame that can be found in a pixel variable
                            "pxl" through pxl.polarization. The data frame should contain the
                            columns "marker", the value_column (e.g. morans_z), and the contrast_column.
    :param target: The label for target components in the contrast_column.
    :param reference: The label for reference components in the contrast_column.
    :param contrast_column: The column to use for the contrast. Defaults to "sample".
    :param cmap: The colormap to use for the heatmap. Defaults to "vlag".
    :param value_column: What polarity metric to use. Defaults to "morans_z".
    :param n_top_pairs: Number of high value marker-pairs to label from positive and negative sides.
    :param min_log_p: marker-pairs only receive a label if -log10 of their p-value is higher than
                      this parameter.

    :return: The figure and axes objects of the plot.
    """
    if isinstance(targets, str):
        targets = [targets]
    elif targets is None:
        targets = polarity_data[contrast_column].unique()
        targets = list(set(targets) - {reference})

    if len(targets) > 5:
        raise ValueError(
            "Only up to 5 targets can be visualized. "
            "Number of requested targets is {len(targets)}."
            "Requested targets are: {targets}."
        )

    differential_polarity = get_differential_polarity(
        polarity_data,
        targets=targets,
        reference=reference,
        contrast_column=contrast_column,
        value_column=value_column,
    )

    fig, axes = plt.subplots(1, len(targets))
    for i, target in enumerate(sorted(targets)):
        ax = axes[i] if len(targets) > 1 else axes
        target_differential_polarity = differential_polarity.loc[
            differential_polarity["target"] == target, :
        ].set_index("marker")
        target_differential_polarity["target_mean"] = (
            polarity_data[polarity_data[contrast_column] == target]
            .groupby("marker")[value_column]
            .mean()
        )

        p = ax.scatter(
            x=target_differential_polarity["median_difference"],
            y=-np.log10(target_differential_polarity["p_adj"]),
            c=target_differential_polarity["target_mean"],
            s=20,
            marker="o",
            cmap=cmap,
        )

        ax.set(
            xlabel="Median difference",
            ylabel=r"$-\log_{10}$(adj. p-value)",
            title=f"Differential polarity\nbetween {target}\nand {reference}",
        )
        ax.title.set_y(1.05)
        fig.colorbar(p, label="Mean target polarity score", cmap=cmap)
        _add_top_marker_labels(
            target_differential_polarity,
            n_top_pairs=n_top_pairs,
            min_log_p=min_log_p,
            ax=ax,
        )
    fig.subplots_adjust(top=0.8)
    fig.set_size_inches(6 * len(targets), 5)
    return fig, axes



"""Code adapted from cutadapt.

Modifications Copyright © 2025 Pixelgen Technologies AB.
Under the same license terms as the original code.

Original copyright notice:
Copyright (c) 2010 Marcel Martin <marcel.martin@scilifelab.se> and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import enum
import io
import logging
import multiprocessing
import queue
import sys
import traceback
from collections import defaultdict
from multiprocessing import Queue
from multiprocessing.connection import Connection
from pathlib import Path
from queue import Empty
from typing import TYPE_CHECKING, Any, Iterator, Optional, Protocol, Tuple, Type

import pyarrow as pa
import pyarrow.parquet as pq
from cutadapt.files import (
    FileFormat,
    InputFiles,
    InputPaths,
)
from cutadapt.report import Statistics
from cutadapt.runners import PipelineRunner
from cutadapt.utils import Progress

from pixelator.pna.demux.barcode_demuxer import BarcodeDemuxer, DemuxRecordBatch
from pixelator.pna.read_processing.runners import ReaderProcess, logger

logger = logging.getLogger()

mpctx = multiprocessing.get_context("spawn")


# See https://github.com/python/typeshed/issues/9860
if TYPE_CHECKING:
    mpctx_Process = multiprocessing.Process  # pragma: no cover
else:
    mpctx_Process = mpctx.Process


class DemuxPipeline:
    """A pipeline that processes reads and sends them to a writer queue for further processing."""

    def __init__(
        self, demuxer: BarcodeDemuxer, writer_queue: Optional[Queue] = None
    ) -> None:
        """Initialize a pipeline that processes reads and sends them to a writer queue for further processing.

        :param demuxer: The barcode demuxer to use
        :param writer_queue: The queue to send the processed reads to
        """
        self._demuxer = demuxer
        self._writer_queue = writer_queue

    def set_writer_queue(self, writer_queue: Queue) -> None:
        """Set the writer queue for the pipeline."""
        self._writer_queue = writer_queue

    def process_reads(
        self,
        infiles: InputFiles,
        progress: Optional[Progress] = None,
    ) -> Tuple[int, int, Optional[int]]:
        """Run the pipeline. Return statistics.

        :param infiles: Input chunks of data as a file-like object
        :param progress: A progress object to update
        """
        assert self._writer_queue is not None

        reader = infiles.open()
        n = 0  # no. of processed reads
        total_bp = 0
        _demuxer = self._demuxer

        for read in reader:
            n += 1
            if n % 10000 == 0 and progress is not None:
                progress.update(10000)

            total_bp += len(read)
            res = _demuxer(read)
            self._submit(res)

        if progress is not None:
            progress.update(n % 10000)

        infiles.close()
        return (n, total_bp, None)

    def flush(self):
        """Flush any buffered data from the pipeline."""
        for res in self._demuxer.flush():
            self._submit(res)

    def close(self):
        """Flush the remaining items to the queue and close the queue.

        Closing the queue and joining the tread managing it
        in the current active worker is needed to make sure
        all data is visible to consumers.
        """
        self.flush()
        self._writer_queue.close()
        self._writer_queue.join_thread()

    def _submit(self, res):
        if not res:
            return

        def _submit_single(r):
            submitted = False
            while not submitted:
                try:
                    self._writer_queue.put(r, timeout=5)
                    submitted = True
                except queue.Full:
                    pass

        if isinstance(res, list):
            for r in res:
                _submit_single(r)
        else:
            _submit_single(res)


class DemuxWriterProcessMessages(enum.IntEnum):
    """Enum for messages between different processes in the pipeline."""

    READY = 0
    DONE = 1
    EXCEPTION = 2
    SHUTDOWN = 3


class RecordBatchList:
    """A small helper class to store RecordBatches in a list and keep track of the total size."""

    def __init__(self):
        """Initialize a RecordBatchList instance."""
        self._batches = []
        self._total_num_rows = 0

    def append(self, batch: pa.RecordBatch):
        """Append a new RecordBatch to the list."""
        self._batches.append(batch)
        self._total_num_rows += batch.num_rows

    def batches(self) -> Iterator[pa.RecordBatch]:
        """Return an iterator over the RecordBatches."""
        return iter(self._batches)

    def clear(self):
        """Clear the list of RecordBatches."""
        self._batches.clear()
        self._total_num_rows = 0

    def num_rows(self):
        """Return the total size of RecordBatches in the list."""
        return self._total_num_rows


class DemuxFilenamePolicy(Protocol):
    """A policy to determine the filename of the output files.

    The filename is determined by the sample name and the group index.
    """

    def get_filename(self, group_index: int) -> str:
        """Return the filename for the given group index.

        :param group_index: The index of the group to write the data to
        """
        ...


class PartsFilenamePolicy(DemuxFilenamePolicy):
    """A filename policy for creating demux output files.

    Each group will be written to a separate file with the
    `part_???.demux.arrow` suffix.
    """

    def __init__(self, prefix: str):
        """Initialize the filename policy.

        :param prefix: The prefix to use for the output files.
        """
        super().__init__()
        self.prefix = prefix

    def get_filename(self, group_index: int) -> str:
        """Return the filename for the given group index."""
        return f"{self.prefix}.part_{group_index:03d}.parquet"


class IndependentMarkersFilenamePolicy(DemuxFilenamePolicy):
    """A filename policy for creating demux output files.

    The flat space of group_index is split into two parts,
    one for marker1 and one for marker2.

    Small marker groups can still be binned together, so the
    number of files is not equal to twice the number of markers
    in a panel.
    """

    def __init__(self, prefix: str, m1_map: dict[str, int], m2_map: dict[str, int]):
        """Initialize the filename policy.

        :param prefix: The prefix to use for the output files.
        """
        super().__init__()
        self.prefix = prefix

        # The set of group indices that are partitions for marker1
        self._m1_groups = set(m1_map.values())
        self._m1_min = min(self._m1_groups)
        # The set of group indices that are partitions for marker2
        self._m2_groups = set(m2_map.values())
        self._m2_min = min(self._m2_groups)

    def _rescaled_m1(self, idx: int):
        return idx - self._m1_min

    def _rescaled_m2(self, idx: int):
        return idx - self._m2_min

    def get_filename(self, group_index: int) -> str:
        """Return the filename for the given group index."""
        if group_index in self._m1_groups:
            idx = self._rescaled_m1(group_index)
            return f"{self.prefix}.m1.part_{idx:03d}.parquet"

        if group_index in self._m2_groups:
            idx = self._rescaled_m2(group_index)
            return f"{self.prefix}.m2.part_{idx:03d}.parquet"

        raise ValueError(f"Invalid group index {group_index}")


class DemuxWriterProcess(mpctx_Process):
    """A custom writer process for streaming data into multiple parquet files."""

    def __init__(
        self,
        output_directory: Path,
        filename_policy: DemuxFilenamePolicy,
        queue: multiprocessing.Queue,
        connection: Connection,
        schema: pa.schema,
    ):
        """Initialize a DemuxWriterProcess instance.

        :param output_directory: The directory to write the output arrow files to
        :param filename_prefix: The prefix to use for the output files
        :param queue: The queue to receive data from the worker processes
        :param connection: The connection to the parent process
        :param schema: The schema of the arrow Tables received from the worker nodes
        """
        super().__init__()

        # self._output_prefix = output_prefix
        #: The input connections to the individual worker processes
        self.output_directory = output_directory
        self.filename_policy = filename_policy
        self._queue = queue
        self._connection = connection
        self._write_buffers: dict[int, RecordBatchList] = defaultdict(RecordBatchList)
        self._batch_size = 10_000
        self._schema = schema

        self._batch_writers: dict[int, pq.ParquetWriter] = {}

        # Write uncompressed here
        # These IPC files will be compressed later to parquet after sorting.
        # Saving them uncompressed makes sure we can mmap the files in the sorting step
        # We save the write_option as a dict since `pa.ipc.IpcWriteOptions` is not picklable.
        self._write_options = {"use_threads": True}

    def open_writer(self, group_index: int):
        """Open a new writer for the given group index."""
        name = self.filename_policy.get_filename(group_index)
        output_file_path = str(self.output_directory / name)
        batch_writer = pq.ParquetWriter(
            output_file_path,
            self._schema,
            compression="zstd",
            compression_level=1,
        )
        self._batch_writers[group_index] = batch_writer
        return batch_writer

    def run(self):
        """Run the writer process."""
        # Give a ready signal to the parent process
        self._connection.send(DemuxWriterProcessMessages.READY)
        _queue = self._queue

        stop_signal = self._connection.poll(timeout=0)
        data = None

        try:
            while not stop_signal:
                try:
                    data = _queue.get(timeout=1)
                    group_index, group = data
                    self.buffer_batch_data(group_index, group)
                except Empty:
                    pass

                # Check for an out-of-band stop signal on the parent connection
                stop_signal = self._connection.poll(timeout=0)

            # Pump the queue empty, all remaining data is already present at this point
            while not _queue.empty():
                data = _queue.get()
                group_index, group = data
                self.buffer_batch_data(group_index, group)

            self.close()
            self._connection.send(DemuxWriterProcessMessages.DONE)
            self._connection.close()
        except Exception as e:
            self.close()
            self._connection.send(DemuxWriterProcessMessages.EXCEPTION)
            self._connection.send((e, traceback.format_exc()))

    def buffer_batch_data(self, group_index, buf: pa.Buffer):
        """Buffer the data from the worker processes in RecordBatches.

        If the buffer is large enough, write it to disk.

        :param group_index: The index of the group to write the data to
        :param buf: The serialized bytes of a RecordBatch received from a worker process
        """
        # Convert the raw buffer back into a RecordBatch
        group_batch = pa.ipc.read_record_batch(buf, self._schema)
        group_data = self._write_buffers[group_index]
        group_data.append(group_batch)

        # Check if we have enough data buffered to write a RecordBatch
        if group_data.num_rows() >= self._batch_size:
            self.write_batches(group_index, group_data)

    def write_batches(self, group_index: int, group_data: RecordBatchList) -> None:
        """Write batches of records to the output file for given `group_index`.

        :param group_index: The index of the group to write the data to
        :param group_data: The RecordBatchList containing the data to write
        """
        table = pa.Table.from_batches(group_data.batches())
        writer = self._batch_writers.get(group_index)

        if writer is None:
            writer = self.open_writer(group_index)

        writer.write_table(table)
        group_data.clear()

    def flush(self) -> None:
        """Flush any remaining data to the writers."""
        for group_index, group_data in self._write_buffers.items():
            if group_data.num_rows() > 0:
                self.write_batches(group_index, group_data)

    def close(self) -> None:
        """Close all writers."""
        self.flush()
        for writer in self._batch_writers.values():
            writer.close()


class DemuxWorkerProcess(mpctx_Process):
    """A custom process that runs the pipeline on chunks of input data.

    The worker process reads data from the read_pipe and processes it using the pipeline.
    The results are sent to the writer_queue for writing to different output files.

    To notify the reader process that it wants data, it puts its own identifier into the
    need_work_queue before attempting to read data from the read_pipe.
    """

    def __init__(
        self,
        id_: int,
        pipeline: DemuxPipeline,
        inpaths: InputPaths,
        read_pipe: Connection,
        parent_pipe: Connection,
        writer_queue: multiprocessing.Queue,
        need_work_queue: multiprocessing.Queue,
        statistics_class: Type[Statistics] = Statistics,
    ) -> None:
        """Initialize the worker process.

        :param id_: The index of the worker
        :param pipeline: The pipeline to run
        :param inpaths: The input paths to read data from
        :param read_pipe: The connection to the reader process
        :param parent_pipe: The connection to the parent process
        :param writer_queue: The queue to send the processed reads to
        :param need_work_queue: The queue to notify the reader process that work is needed
        """
        super().__init__()
        self._id = id_
        self._pipeline = pipeline
        self._n_input_files = len(inpaths.paths)
        self._interleaved_input = inpaths.interleaved
        self._read_pipe = read_pipe
        self._parent_pipe = parent_pipe
        self._writer_queue = writer_queue
        self._need_work_queue = need_work_queue
        self._statistics_class = statistics_class
        self._pipeline.set_writer_queue(self._writer_queue)

    def run(self):
        """Run the demux worker process."""
        try:
            stats = self._statistics_class()
            stats.paired = False

            while True:
                # Notify reader that we need data
                self._need_work_queue.put(self._id)
                chunk_index = self._read_pipe.recv()
                if chunk_index == -1:
                    # reader is done
                    break
                elif chunk_index == -2:
                    # An exception has occurred in the reader
                    e, tb_str = self._read_pipe.recv()
                    logger.error("%s", tb_str)
                    raise e

                # Receive blocks of data from the reader
                files = [
                    io.BytesIO(self._read_pipe.recv_bytes())
                    for _ in range(self._n_input_files)
                ]
                # Pass the data to the pipeline as a file like object
                infiles = InputFiles(*files, interleaved=self._interleaved_input)
                (n, bp1, bp2) = self._pipeline.process_reads(infiles)
                stats += self._statistics_class().collect(n, bp1, bp2, [], [])
                self._send_outfiles(chunk_index, n)

            self._pipeline.close()
            self._parent_pipe.send(-1)
            self._parent_pipe.send(stats)

        except Exception as e:
            self._pipeline.close()

            # Notify the parent process that an exception has occurred
            self._parent_pipe.send(-2)
            self._parent_pipe.send((e, traceback.format_exc()))

    def _send_outfiles(self, chunk_index: int, n_reads: int):
        self._parent_pipe.send(chunk_index)
        self._parent_pipe.send(n_reads)


class WorkerException(Exception):
    """Wrap an exception that occurred in a worker process.

    Capture the traceback string of the original exception.
    This is class is needed because traceback objects are not picklable.
    To communicate the traceback in a multiprocessing setting, we need to
    capture the traceback as a string.

    :ivar e: The original exception
    :ivar tb_str: The traceback string
    """

    def __init__(self, wrapped_exception, tb_str):
        """Initialize the WorkerException object.

        :param wrapped_exception: The original exception
        :param tb_str: The traceback string
        """
        self.e = wrapped_exception
        self.tb_str = tb_str

    def __str__(self):
        """Return a string representation of wrapped the exception."""
        return f"{self.e}\nwith traceback:\n{self.tb_str}"


class ParallelDemuxPipelineRunner(PipelineRunner):
    """Run a Demux pipeline and dispatch work to worker threads.

    - At construction, a reader process is spawned.
    - When run() is called, as many worker processes as requested are spawned.
    - The workers return a tuple of (file_index, data) to the main process.
    - The writer process buffers writes to the output files and writes them in the correct order.
    - write

    If a worker needs work, it puts its own index into a Queue() (_need_work_queue).
    The reader process listens on this queue and sends the raw data to the
    worker that has requested work. For sending the data from reader to worker,
    a Connection() is used. There is one such connection for each worker (self._pipes).

    For sending the processed data from the worker to the main process, there
    is a second set of connections, again one for each worker.

    When the reader is finished, it sends 'poison pills' to all workers.
    When a worker receives this, it sends a poison pill to the main process,
    followed by a Statistics object that contains statistics about all the reads
    processed by that worker.
    """

    def __init__(
        self,
        inpaths: InputPaths,
        n_workers: int,
        output_directory: Path,
        filename_policy: DemuxFilenamePolicy,
        buffer_size: Optional[int] = None,
    ):
        """Create a new ParallelDemuxPipelineRunner instance.

        Args:
            inpaths: The input paths to read data from.
            n_workers: The number of worker processes to spawn.
            output_directory: The directory to write the output files to.
            filename_policy: The policy to determine the filename of the output files.
            buffer_size: The size of the buffer to use for reading data.

        """
        self._n_workers = n_workers
        self._need_work_queue: multiprocessing.Queue = mpctx.Queue()
        self._buffer_size = 4 * 1024**2 if buffer_size is None else buffer_size
        self._inpaths = inpaths
        self._filename_policy = filename_policy

        if not output_directory.exists():
            raise NotADirectoryError(output_directory)

        output_directory.mkdir(parents=True, exist_ok=True)

        # the workers read from these connections
        in_connections = [mpctx.Pipe(duplex=False) for _ in range(self._n_workers)]
        self._worker_input_connection, connw = zip(*in_connections)

        # # Full-duplex pipe for communicating worker output with the writer process
        self._writer_queue = mpctx.Queue(maxsize=self._n_workers * 2)

        try:
            fileno = sys.stdin.fileno()
        except io.UnsupportedOperation:
            # This happens during tests: pytest sets sys.stdin to an object
            # that does not have a file descriptor.
            fileno = -1

        file_format_connection_r, file_format_connection_w = mpctx.Pipe(duplex=False)

        self._reader_process = ReaderProcess(
            *inpaths.paths,
            file_format_connection=file_format_connection_w,
            connections=connw,
            queue=self._need_work_queue,
            buffer_size=self._buffer_size,
            stdin_fd=fileno,
        )
        self._reader_process.daemon = True
        self._reader_process.start()

        self._writer_conn, writer_conn_child = mpctx.Pipe(duplex=True)

        self._writer_process = DemuxWriterProcess(
            output_directory=output_directory,
            filename_policy=self._filename_policy,
            queue=self._writer_queue,
            connection=writer_conn_child,
            schema=DemuxRecordBatch.schema(),
        )
        self._writer_process.daemon = True
        self._writer_process.start()

        self._input_file_format = self._try_receive(file_format_connection_r)

    def _start_workers(
        self, pipeline
    ) -> tuple[list[DemuxWorkerProcess], list[Connection]]:
        """Start the worker processes and return them and the connections to them."""
        workers = []
        out_connection_parent: list[Connection] = []

        for index in range(self._n_workers):
            conn_parent, conn_child = mpctx.Pipe(duplex=True)

            out_connection_parent.append(conn_parent)

            worker = DemuxWorkerProcess(
                index,
                pipeline,
                self._inpaths,
                read_pipe=self._worker_input_connection[index],
                parent_pipe=conn_child,
                writer_queue=self._writer_queue,
                need_work_queue=self._need_work_queue,
            )
            worker.daemon = True
            worker.start()
            workers.append(worker)

        return workers, out_connection_parent

    def run(self, pipeline, progress) -> Statistics:
        """Run a pipeline on the pipeline runner.

        :param pipeline: The pipeline to run.
        :param progress: A progress object to update with the number of processed reads.
        :return: A statistics object.
        """
        workers, connections = self._start_workers(pipeline)

        # Wait for the writer process to signal it is ready to start receiving data
        writer_message = self._writer_conn.recv()
        if writer_message != DemuxWriterProcessMessages.READY:
            raise RuntimeError("Writer process did not signal ready")

        stats = Statistics()
        stats.paired = False

        while connections:
            # Regularly check if the writer process has a message
            writer_message = self._writer_conn.poll(timeout=0)

            if writer_message:
                status = self._writer_conn.recv()

                if status == DemuxWriterProcessMessages.DONE:
                    logger.info("Demux writer process finished")
                    self._writer_process.join()

                if status == DemuxWriterProcessMessages.EXCEPTION:
                    e, tb_str = self._writer_conn.recv()
                    logger.error("An exception occurred in the writer process")
                    logger.error("%s", tb_str)
                    for c in mpctx.active_children():
                        c.terminate()
                    raise e

            # Wait for any of the worker to be signal they are ready to receive work
            # We use a timeout to avoid blocking indefinitely so that we can check
            # for messages on the writer process as well
            try:
                ready_connections: list[Any] = multiprocessing.connection.wait(
                    connections, timeout=1
                )
            except TimeoutError:
                continue

            for connection in ready_connections:
                chunk_index: int = self._try_receive(connection)
                if chunk_index == -1:
                    # the worker is done, the second item send is the statistics
                    cur_stats = self._try_receive(connection)
                    stats += cur_stats
                    connections.remove(connection)
                    continue

                number_of_reads: int = self._try_receive(connection)
                progress.update(number_of_reads)

        for w in workers:
            w.join()

        self._reader_process.join()

        # Signal the writer process to stop
        self._writer_conn.send(-1)
        self._writer_process.join()
        progress.close()
        return stats

    @staticmethod
    def _try_receive_no_wait(connection):
        has_results = connection.poll(timeout=1)

        if not has_results:
            return None

        result = connection.recv()
        if result == -2:
            # An exception has occurred on the other end
            e, tb_str = connection.recv()
            # The other end does not send an actual traceback object because these are
            # not picklable, but a string representation.
            logger.debug("%s", tb_str)
            for child in multiprocessing.active_children():
                child.terminate()
            raise WorkerException(e, tb_str)

        return result

    @staticmethod
    def _try_receive(connection):
        """Try to receive data over `connection` and return it.

        If an exception was received, raise it.
        """
        result = connection.recv()
        if result == -2:
            # An exception has occurred on the other end
            e, tb_str = connection.recv()
            # The other end does not send an actual traceback object because these are
            # not picklable, but a string representation.
            logger.debug("%s", tb_str)
            for child in multiprocessing.active_children():
                child.terminate()
            raise WorkerException(e, tb_str)

        return result

    def close(self) -> None:
        """Close the pipeline runner."""
        pass

    def input_file_format(self) -> FileFormat:
        """Return the file format of the input files."""
        return self._input_file_format



"""Functions to create an interactive qc report.

Copyright © 2022 Pixelgen Technologies AB.
"""

from .builder import QCReportBuilder
from .collect import collect_report_data
from .main import create_dynamic_report, make_report
from .types import InfoAndMetrics, Metrics, QCReportData, SampleInfo

__all__ = [
    "QCReportData",
    "InfoAndMetrics",
    "Metrics",
    "SampleInfo",
    "collect_report_data",
    "QCReportBuilder",
    "create_dynamic_report",
    "make_report",
]



"""Copyright © 2023 Pixelgen Technologies AB."""

from pathlib import Path
from typing import Any

from ruamel import yaml

from pixelator.types import PathType


def load_yaml_file(path: PathType) -> Any:
    """Load an arbitrary yaml file.

    :param path: path to the yaml file
    :raises FileExistsError: If the path does not exist
    :raises TypeError: If the path is not a yaml file
    :returns: a yaml object
    """
    path = Path(path)
    if not path.is_file():
        raise FileExistsError(f"{path} is not a file")

    if not path.suffix == ".yaml":
        raise TypeError(f"{path} is not a yaml file")

    yaml_loader = yaml.YAML(typ="safe")
    with open(path, "r") as cf:
        data = yaml_loader.load(cf)

    return data



"""Implementation of the pixelator Graph protocol based on networkx.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
import warnings
from timeit import default_timer as timer
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    get_args,
)

import networkx as nx
import numpy as np
import pandas as pd
import polars as pl
import scipy as sp
from networkx.algorithms import bipartite as nx_bipartite
from scipy.sparse import csr_matrix

from pixelator.common.graph.backends.protocol import (
    Edge,
    EdgeSequence,
    GraphBackend,
    SupportedLayoutAlgorithm,
    Vertex,
    VertexClustering,
    VertexSequence,
)

if TYPE_CHECKING:
    from pixelator.mpx.graph import Graph

logger = logging.getLogger(__name__)


class NetworkXGraphBackend(GraphBackend):
    """`NetworkXGraphBackend` represents a graph, using networkx."""

    def __init__(
        self,
        raw: Optional[nx.Graph] = None,
    ):
        """Create a new Graph instance.

        Create a Graph instance (as an end-user this is probably not the interface
        you are looking for). Try `Graph.from_edgelist`.

        :param raw: The underlying raw representation of the graph, defaults to None
        """
        self._raw = raw

    @staticmethod
    def _build_plain_graph_from_edgelist(
        df: pl.DataFrame,
        create_using: Union[nx.Graph, nx.MultiGraph],
    ) -> Union[nx.Graph, nx.MultiGraph]:
        g = nx.empty_graph(0, create_using)

        # TODO Look at how to deal with setting project_pushdown=False
        # here. If it is needed or not seems to depend on the
        # exact call context, so it might be that we can actually
        # enable it again here and improve the memory usage.
        for idx, row in enumerate(df.iter_rows(named=False, buffer_size=1000)):
            g.add_edge(row[0], row[1], index=idx)
        return g

    @staticmethod
    def _build_graph_with_node_counts_from_edgelist(
        df: pl.DataFrame,
        create_using: Union[nx.Graph, nx.MultiGraph],
    ) -> Union[nx.Graph, nx.MultiGraph]:
        unique_markers = set(df.select("marker").unique()["marker"].to_list())
        initial_marker_dict = {marker: 0 for marker in unique_markers}

        g: nx.Graph = nx.empty_graph(0, create_using)

        for idx, row in enumerate(
            (
                df.select(["upia", "upib", "marker"]).iter_rows(
                    named=False, buffer_size=1000
                )
            )
        ):
            # We are duplicating code here, since it gives
            # a performance boost of roughly 10% here.
            # Which isn't that much, but will add up when
            # we have large edge lists.
            node_1 = row[0]
            node_2 = row[1]
            marker = row[2]

            existing_node_1 = g.nodes.get(node_1)
            if existing_node_1:
                existing_node_1["markers"][marker] += 1
            else:
                marker_dict = initial_marker_dict.copy()
                marker_dict[marker] += 1
                g.add_node(node_1, markers=marker_dict)

            existing_node_2 = g.nodes.get(node_2)
            if existing_node_2:
                existing_node_2["markers"][marker] += 1
            else:
                marker_dict = initial_marker_dict.copy()
                marker_dict[marker] += 1
                g.add_node(node_2, markers=marker_dict)

            g.add_edge(node_1, node_2, index=idx)

        return g

    @staticmethod
    def _add_node_attributes(
        graph: Union[nx.Graph, nx.MultiGraph], a_nodes: set[str]
    ) -> None:
        node_names = {node: node for node in graph.nodes()}
        pixel_type = {node: "A" if node in a_nodes else "B" for node in graph.nodes()}
        type_ = {node: node in a_nodes for node in graph.nodes()}
        nx.set_node_attributes(graph, node_names, "name")
        nx.set_node_attributes(graph, pixel_type, "pixel_type")
        nx.set_node_attributes(graph, type_, "type")

    @staticmethod
    def _project_on_a_nodes(
        graph: Union[nx.Graph, nx.MultiGraph], a_nodes: set[str]
    ) -> Union[nx.Graph, nx.MultiGraph]:
        if isinstance(graph, nx.MultiGraph):
            warnings.warn(
                "Using `use_full_bipartite=False` together with `simplify=False` "
                "will still implicitly simplify the graph, since the multi-edges "
                "will be lost upon A-node projection."
            )
            graph = nx.Graph(graph)

        return nx_bipartite.projected_graph(graph, a_nodes)

    @staticmethod
    def _build_graph_with_marker_counts(
        edgelist: pl.LazyFrame, simplify: bool, use_full_bipartite: bool
    ) -> Union[nx.Graph, nx.MultiGraph]:
        edgelist_cl = edgelist.select(["upia", "upib", "marker"]).collect()
        graph = NetworkXGraphBackend._build_graph_with_node_counts_from_edgelist(
            edgelist_cl,
            create_using=nx.Graph if simplify else nx.MultiGraph,
        )
        a_nodes = set(edgelist_cl.select(["upia"]).unique()["upia"].to_list())
        NetworkXGraphBackend._add_node_attributes(graph, a_nodes)
        if use_full_bipartite:
            return graph
        return NetworkXGraphBackend._project_on_a_nodes(graph, a_nodes)

    @staticmethod
    def _build_plain_graph(
        edgelist: pl.LazyFrame, simplify: bool, use_full_bipartite: bool
    ) -> Union[nx.Graph, nx.MultiGraph]:
        edgelist_cl = edgelist.select(["upia", "upib"]).collect()
        graph = NetworkXGraphBackend._build_plain_graph_from_edgelist(
            edgelist_cl,
            create_using=nx.Graph if simplify else nx.MultiGraph,
        )
        a_nodes = set(edgelist_cl.select(["upia"]).unique()["upia"].to_list())
        NetworkXGraphBackend._add_node_attributes(graph, a_nodes)
        if use_full_bipartite:
            return graph
        return NetworkXGraphBackend._project_on_a_nodes(graph, a_nodes)

    @staticmethod
    def from_edgelist(
        edgelist: Union[pd.DataFrame, pl.LazyFrame],
        add_marker_counts: bool,
        simplify: bool,
        use_full_bipartite: bool,
        convert_indices_to_integers: bool = True,
    ) -> NetworkXGraphBackend:
        """Build a graph from an edgelist.

        Build a Graph from an edge list (pd.DataFrame). Multiple options are available
        to build the graph, `add_marker_counts` will add a dictionary of marker counts
        to each node, `simplify` will remove redundant edges and `use_full_bipartite`
        will not project the graph (UPIA).

        The graph will contain the edge attributes present in the edge list when
        `use_full_bipartite` is True and a dictionary of marker counts in each
        vertex (node) when `add_marker_counts` is True. If `use_full_bipartite` is
        False or `simplify` is True the edge attributes will be lost.

        :param edgelist: the edge list (dataframe) corresponding to the graph
        :param add_marker_counts: add a dictionary of marker counts to each node
        :param simplify: simplifies the graph (remove redundant edges)
        :param use_full_bipartite: use the bipartite graph instead of the projection
                                  (UPIA)
        :param convert_indices_to_integers: convert the indices to integers (this is the default)
        :returns: a GraphBackend instance
        :rtype: NetworkXGraphBackend
        :raises: AssertionError when the input edge list is not valid
        """
        if isinstance(edgelist, pd.DataFrame):
            edgelist: pl.LazyFrame = pl.LazyFrame(edgelist)  # type: ignore

        if add_marker_counts:
            graph = NetworkXGraphBackend._build_graph_with_marker_counts(
                edgelist, simplify, use_full_bipartite
            )
        else:
            graph = NetworkXGraphBackend._build_plain_graph(
                edgelist, simplify, use_full_bipartite
            )

        # TODO igraph uses integer indexing. This converts the networkx graph to using
        # the same-ish schema. We probably evaluate if this is really necessary later,
        # or potentially only do it on request.
        if convert_indices_to_integers:
            graph = nx.convert_node_labels_to_integers(graph, ordering="sorted")

        return NetworkXGraphBackend(raw=graph)

    @staticmethod
    def from_raw(graph: Union[nx.Graph, nx.MultiGraph]) -> NetworkXGraphBackend:
        """Generate a Graph from an networkx.Graph object.

        :param graph: input networkx graph to use
        :return: A pixelator Graph object
        :rtype: NetworkXGraphBackend
        """
        return NetworkXGraphBackend(graph)

    @property
    def raw(self):
        """Get the raw underlying graph representation."""
        return self._raw

    @property
    def vs(self):
        """Get a sequence of the vertices in the Graph instance."""
        return NetworkxBasedVertexSequence(
            vertices=[
                NetworkxBasedVertex(v[0], v[1], self.raw)
                for v in self.raw.nodes(data=True)
            ]
        )

    @property
    def es(self):
        """A sequence of the edges in the Graph instance."""
        return NetworkxBasedEdgeSequence(self._raw, self._raw.edges(data=True))

    def vcount(self):
        """Get the total number of vertices in the Graph instance."""
        return self._raw.number_of_nodes()

    def ecount(self):
        """Get the total number of edges in the Graph instance."""
        return self._raw.number_of_edges()

    def get_adjacency_sparse(
        self, node_ordering: Iterable[Any] | None = None
    ) -> csr_matrix:
        """Get the sparse adjacency matrix.

        :param node_ordering: Control the node ordering in the adjacency matrix
        :return: a sparse adjacency matrix
        """
        return nx.to_scipy_sparse_array(self._raw, nodelist=node_ordering)

    def connected_components(self) -> NetworkxBasedVertexClustering:
        """Get the connected components in the Graph instance."""
        return NetworkxBasedVertexClustering(
            self._raw, nx.connected_components(self._raw)
        )

    def _layout_coordinates(
        self,
        layout_algorithm: SupportedLayoutAlgorithm = "wpmds_3d",
        random_seed: Optional[int] = None,
        **kwargs,
    ) -> pd.DataFrame:
        if layout_algorithm not in get_args(SupportedLayoutAlgorithm):
            raise AssertionError(
                (
                    f"{layout_algorithm} not allowed `layout_algorithm` option. "
                    f"Options are: {'/'.join(get_args(SupportedLayoutAlgorithm))}"
                )
            )

        if not self._raw:
            raise ValueError("Trying to get layout for empty Graph instance.")
        raw = self._raw  # type: nx.Graph

        if layout_algorithm == "kamada_kawai":
            layout_inst = nx.kamada_kawai_layout(
                raw, pos=nx.random_layout(raw, seed=random_seed)
            )
        if layout_algorithm == "kamada_kawai_3d":
            layout_inst = nx.kamada_kawai_layout(
                raw, pos=nx.random_layout(raw, seed=random_seed, dim=3), dim=3
            )
        if layout_algorithm == "fruchterman_reingold":
            layout_inst = nx.spring_layout(raw, seed=random_seed)
        if layout_algorithm == "fruchterman_reingold_3d":
            layout_inst = nx.spring_layout(raw, dim=3, seed=random_seed)
        if layout_algorithm == "pmds":
            layout_inst = pmds_layout(raw, seed=random_seed, **kwargs)
        if layout_algorithm == "pmds_3d":
            layout_inst = pmds_layout(raw, dim=3, seed=random_seed, **kwargs)
        if layout_algorithm == "wpmds_3d":
            layout_inst = pmds_layout(
                raw, dim=3, weights="prob_dist", seed=random_seed, **kwargs
            )

        coordinates = pd.DataFrame.from_dict(
            layout_inst,
            orient="index",
            columns=["x", "y", "z"] if "3d" in layout_algorithm else ["x", "y"],
        )

        return coordinates

    @staticmethod
    def _normalize_to_unit_sphere(coordinates):
        coordinates[["x_norm", "y_norm", "z_norm"]] = (
            coordinates[["x", "y", "z"]]
            / (1 * np.linalg.norm(np.asarray(coordinates), axis=1))[:, None]
        )
        return coordinates

    def node_marker_counts(self) -> pd.DataFrame:
        """Get the marker counts of each node as a dataframe.

        :return: node markers as a dataframe
        :rtype: pd.DataFrame
        :raises: AssertionError if graph nodes don't include markers
        """
        if "markers" not in self.vs.attributes():
            raise AssertionError("Could not find 'markers' in vertex attributes")
        markers = list(sorted(next(iter(self.vs))["markers"].keys()))
        node_marker_counts = pd.DataFrame.from_records(
            list(self.vs.get_attribute("markers")),
            columns=markers,
            index=list(self.raw.nodes),
        )
        node_marker_counts = node_marker_counts.reindex(
            sorted(node_marker_counts.columns), axis=1
        )
        node_marker_counts.columns.name = "markers"
        node_marker_counts.columns = node_marker_counts.columns.astype(
            "string[pyarrow]"
        )
        node_marker_counts.index.name = "node"
        return node_marker_counts

    def layout_coordinates(
        self,
        layout_algorithm: SupportedLayoutAlgorithm = "wpmds_3d",
        only_keep_a_pixels: bool = True,
        get_node_marker_matrix: bool = True,
        random_seed: Optional[int] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Generate coordinates and (optionally) node marker counts for plotting.

        Generate a dataframe with coordinates, and (optionally) node marker
        counts to use that can be used for plotting.

        The layout options are:
          - pmds
          - pmds_3d
          - fruchterman_reingold
          - fruchterman_reingold_3d
          - kamada_kawai
          - kamada_kawai_3d
          - wpmds_3d

        For most cases the `pmds` options should be about 10-100x faster
        than the force directed layout methods, i.e. `fruchterman_reingold`
        and `kamada_kawai`. Among the force directed layout methods,
        `fruchterman_reingold` is generally faster than `kamada_kawai`. The
        `wpmds_3d` method uses edge weights to improve the layout, but is slightly
        slower than `pmds_3d`.

        :param layout_algorithm: the layout algorithm to use to generate the coordinates
        :param only_keep_a_pixels: If true, only keep the a-pixels
        :param get_node_marker_matrix: Add a matrix of marker counts to each
                                       node if True.
        :param random_seed: used as the seed for graph layouts with a stochastic
                            element. Useful to get deterministic layouts across
                            method calls.
        :param **kwargs: will be passed to the underlying layout implementation
        :return: the coordinates and markers (if activated) as a dataframe
        :rtype: pd.DataFrame
        :raises: AssertionError if the provided `layout_algorithm` is not valid
        :raises: ValueError if the provided current graph instance is empty
        """
        start_time = timer()
        coordinates = self._layout_coordinates(
            layout_algorithm=layout_algorithm, random_seed=random_seed, **kwargs
        )

        # If we are doing a 3D layout we add the option of normalized
        # vectors where we scale the length of each point vector to be one, so that
        # we have the option of doing a spherical projection of the graph
        if len(coordinates.columns) == 3:
            coordinates = self._normalize_to_unit_sphere(coordinates)

        if get_node_marker_matrix:
            node_marker_counts = self.node_marker_counts()  # type: ignore
            df = pd.concat([coordinates, node_marker_counts], axis=1)
        else:
            df = coordinates
        pixel_name_and_type = pd.DataFrame.from_records(
            [(v.index, v["name"], v["pixel_type"]) for v in self.vs.vertices()],
            columns=["index", "name", "pixel_type"],
        )
        pixel_name_and_type = pixel_name_and_type.set_index("index", drop=False)
        df = pd.concat([pixel_name_and_type, df], axis=1)

        if only_keep_a_pixels:
            df = df[df["pixel_type"] == "A"]

        logger.debug(
            "Layout computation using %s took %.2f seconds",
            layout_algorithm,
            timer() - start_time,
        )
        return df

    def get_edge_dataframe(self):
        """Get the edges as a pandas DataFrame."""
        return nx.to_pandas_edgelist(self.raw)

    def get_vertex_dataframe(self):
        """Get all vertices as a pandas DataFrame."""
        return pd.DataFrame(list(self.raw.nodes()), columns=["name"])

    def add_edges(self, edges: Iterable[Tuple[int]]) -> None:  # noqa: DOC501
        """Add edges to the graph instance.

        :param edges: Add the following edges to the graph instance.
        """
        self.raw.add_edges_from(edges)

    def add_vertices(self, n_vertices: int, attrs: Dict[str, List]) -> None:
        """Add some number of vertices to the graph instance.

        :param n_vertices: the number of vertices to be added to the graph instance.
        :param attrs: dict of sequences, all of length equal to the number of vertices
                      to be added, containing the attributes of the new vertices. If
                      `n_vertices=1` then they have to be lists of length 1.
        :raises IndexError: if the number of graph vertices to add and lists of
                            attributes are of different lengths
        """
        raise NotImplementedError()

    def add_names_to_vertexes(self, vs_names: List[str]) -> None:
        """Rename the current vertices on the graph instance.

        :param vs_names: Add the following vertices to the graph instance.
        :raises ValueError: if the graph is empty
        :raises IndexError: if the number of graph vertices and list of names are
                            of different length
        """
        raise NotImplementedError()


class NetworkxBasedVertex(Vertex):
    """A Vertex instance that plays well with NetworkX."""

    def __init__(
        self, index: int | str, data: Dict, graph: Union[nx.Graph, nx.MultiGraph]
    ):
        """Create a new NetworkxBasedVertex instance."""
        self._index = index
        self._data = data
        self._graph = graph

    @property
    def index(self):
        """Get the index of the vertex."""
        return self._index

    @property
    def data(self) -> Dict:
        """Get the data of the vertex as a dict."""
        return self._data

    def __getitem__(self, attr: str) -> Any:
        """Get the attr of the provided vertex."""
        return self._data[attr]

    def __setitem__(self, attr: str, value: Any) -> None:
        """Set the attr of the vertex."""
        self._data[attr] = value

    def neighbors(self) -> VertexSequence:
        """Get the neighbors of the vertex."""
        neighbor_vertices = set(self._graph.neighbors(self._index))

        def generate_neighbors():
            for node_idx, data in self._graph.nodes(data=True):
                if node_idx in neighbor_vertices:
                    yield NetworkxBasedVertex(node_idx, data, self._graph)

        return NetworkxBasedVertexSequence(generate_neighbors())


class NetworkxBasedEdge(Edge):
    """An Edge instance backed by a Networkx Edge."""

    def __init__(
        self,
        edge_tuple: Tuple[int | str, int | str, Any],
        graph: Union[nx.Graph, nx.MultiGraph],
    ):
        """Create a NetworkxBasedEdge instance."""
        self.edge_tuple = edge_tuple
        self._graph = graph

    def __eq__(self, other: object) -> bool:
        """Determine equality of edge and `other`."""
        if isinstance(other, NetworkxBasedEdge):
            return self.index == other.index
        return False

    def __hash__(self) -> int:
        """Compute the hash of the edge."""
        return hash(self.index)

    @property
    def index(self) -> int:
        """The index of the edge."""
        return self.edge_tuple[2]["index"]

    @property
    def vertex_tuple(self) -> Tuple[Vertex, Vertex]:
        """Return the vertices the edge connects as a tuple."""
        v1_idx, v2_idx, _ = self.edge_tuple
        node_1_data = self._graph.nodes[v1_idx]
        node_2_data = self._graph.nodes[v2_idx]
        return (
            NetworkxBasedVertex(v1_idx, node_1_data, graph=self._graph),
            NetworkxBasedVertex(v2_idx, node_2_data, graph=self._graph),
        )


class NetworkxBasedVertexSequence(VertexSequence):
    """Proxy for a networkx based vertex sequence."""

    def __init__(self, vertices: Iterable[NetworkxBasedVertex]) -> None:
        """Instantiate a new NetworkxBasedVertexSequence."""
        self._vertices: Dict[int, NetworkxBasedVertex] = {v.index: v for v in vertices}

    def __len__(self) -> int:
        """Get the number of vertexes."""
        return len(self._vertices.keys())

    def vertices(self) -> Iterable[Vertex]:
        """Get an iterable of vertices."""
        return self._vertices.values()

    def __iter__(self) -> Iterator[Vertex]:
        """Get an iterator over the vertices in the sequence."""
        return iter(self._vertices.values())

    def attributes(self) -> Set[str]:
        """Get all attributes associated with the vertices."""

        def all_attributes():
            for node in self._vertices.values():
                for key in node.data.keys():
                    yield key

        return set(all_attributes())

    def get_vertex(self, vertex_id: int) -> Vertex:
        """Get the Vertex corresponding to the vertex id."""
        return self._vertices[vertex_id]

    def get_attribute(self, attr: str) -> Iterable[Any]:
        """Get the values of the attribute."""
        for node in self._vertices.values():
            yield node.data[attr]

    def select_where(self, key, value) -> VertexSequence:
        """Select a subset of vertices."""
        return NetworkxBasedVertexSequence(
            [
                NetworkxBasedVertex(idx, vertex.data, vertex._graph)
                for idx, vertex in self._vertices.items()
                if vertex.data[key] == value
            ],
        )


class NetworkxBasedEdgeSequence(EdgeSequence):
    """Proxy for a networkx based edge sequence."""

    def __init__(
        self, graph: Union[nx.Graph, nx.MultiGraph], edges: Iterable[NetworkxBasedEdge]
    ) -> None:
        """Instantiate a new NetworkxBasedEdgeSequence."""
        self._graph = graph
        self._edges = edges

    def __len__(self) -> int:
        """Get the number of edges."""
        return len(list(iter(self)))

    def __iter__(self) -> Iterator[Edge]:
        """Get an iterator over the edges."""
        for edge in self._edges:
            yield edge

    def select_where(self, key, value) -> EdgeSequence:
        """Select a subset of edges."""
        return NetworkxBasedEdgeSequence(
            self._graph,
            [
                NetworkxBasedEdge((v1, v2, data), self._graph)
                for v1, v2, data in self._graph.edges(data=True)
                if data[key] == value
            ],
        )

    def select_within(self, component) -> EdgeSequence:
        """Select a subset of edges."""
        return NetworkxBasedEdgeSequence(
            self._graph,
            [
                NetworkxBasedEdge(edge, self._graph)
                for edge in self._graph.subgraph(component).edges(data=True)
            ],
        )


class NetworkxBasedVertexClustering(VertexClustering):
    """Wrapper class for a cluster of vertexes."""

    def __init__(
        self, graph: Union[nx.Graph, nx.MultiGraph], clustering: Iterable[Set]
    ) -> None:
        """Instantiate a NetworkxBasedVertexClustering."""
        self._graph = graph
        self._clustering = list(clustering)

    def __len__(self) -> int:
        """Get the number of clusters."""
        return len(self._clustering)

    def __iter__(self) -> Iterator[NetworkxBasedVertexSequence]:
        """Provide an iterator over the clusters."""
        for cluster in self._clustering:
            yield NetworkxBasedVertexSequence(
                [
                    NetworkxBasedVertex(v[0], v[1], self._graph)
                    for v in self._graph.subgraph(cluster).nodes(data=True)
                ]
            )

    @property
    def modularity(self) -> float:
        """Get the modularity of the clusters."""
        return nx.community.modularity(self._graph, communities=self._clustering)

    def _crossing_edges(self):
        def find_crossing_edges():
            graph = self._graph
            for cluster in self._clustering:
                # Setting `nbunch2=None` means that we look for all edges
                # that go out of that cluster.
                for crossing_edge in nx.edge_boundary(
                    graph, nbunch1=cluster, nbunch2=None, data=True
                ):
                    yield NetworkxBasedEdge(crossing_edge, self._graph)

        return {e for e in find_crossing_edges()}

    def crossing(self) -> EdgeSequence:
        """Get any crossing edges."""
        return NetworkxBasedEdgeSequence(self._graph, self._crossing_edges())

    def giant(self) -> Graph:
        """Get the largest component."""
        from pixelator.mpx.graph import Graph

        return Graph(
            NetworkXGraphBackend(
                self._graph.subgraph(max(self._clustering, key=len).copy())
            )
        )

    def subgraphs(self) -> Iterable[Graph]:
        """Get subgraphs of each cluster."""
        from pixelator.mpx.graph import Graph

        return [
            Graph(NetworkXGraphBackend(self._graph.subgraph(cluster).copy()))
            for cluster in self._clustering
        ]


def pmds_layout(
    g: nx.Graph,
    pivots: int = 200,
    dim: int = 2,
    weights: Optional[Union[np.ndarray, str]] = None,
    seed: Optional[int] = None,
) -> pd.DataFrame:
    """Calculate a pivot MDS layout for a graph as described in [1]_.

    The algorithm is similar to classical multidimensional scaling (MDS),
    but uses only a smalls set of random pivot nodes. The algorithm is
    considerably faster than MDS and therefore scales better to large graphs.
    The topology of resulting layouts are deterministic for a given seed,
    but may be mirrored across different systems due to variations in
    floating-point precision.

    .. [1] Brandes U, Pich C. Eigensolver Methods for Progressive
        Multidimensional Scaling of Large Data. International Symposium
        on Graph Drawing, 2007. Lecture Notes in Computer Science, vol
        4372. doi: 10.1007/978-3-540-70904-6_6.

    :param g: A networkx graph object
    :param pivots: The number of pivot nodes to use
    :param dim: The dimension of the layout
    :param weights: Edge weights to use for the layout computation.
    Options are:

    * an np.array with non-negative values (same number of elements as edges in g)
    * "prob_dist" to weight each edge (i, j) by -log(P)^3, where P is the probability
        of a random walker to go from i to j in 5 steps and then back again (j->i)
        in 5 steps. For this computation, self-loops are added to the graph to ensure
        that all transitions are possible.
    * None to use unweighted shortest path lengths
    :param seed: Set seed for reproducibility
    :return: A dataframe with layout coordinates
    :rtype: pd.DataFrame
    :raises ValueError: if conditions are not met
    """
    if not nx.is_connected(g):
        raise ValueError("Only connected graphs are supported.")

    n_nodes = len(g.nodes)
    if pivots >= n_nodes:
        total_nodes = n_nodes
        warnings.warn(
            f"'pivots' ({pivots}) should be less than the number of "
            f"nodes ({total_nodes}) in the graph. Using all nodes as 'pivots'."
        )
        pivots = total_nodes

    if dim not in [2, 3]:
        raise ValueError("'dim' must be either 2 or 3.")

    if pivots < dim:
        raise ValueError("'pivots' must be greater than or equal to dim.")

    if n_nodes <= dim:
        raise ValueError(
            f"Number of nodes in the graph ({n_nodes}) must be greater than or equal to 'dim' ({dim})."
        )

    pivot_lower_bound = np.min([np.floor(0.2 * len(g.nodes)), 50])
    if pivots < pivot_lower_bound:
        raise ValueError(
            f"'pivots' must be greater than or equal to {pivot_lower_bound}"
        )

    if isinstance(weights, str):
        if weights != "prob_dist":
            raise ValueError("If 'weights' is a string, it must be 'prob_dist'.")
        weights = -(np.log(_prob_edge_weights(g, k=5)) ** 3)
    elif isinstance(weights, np.ndarray):
        if len(weights) != len(g.edges):
            raise ValueError(
                "'weights' must have the same length as the number of edges in the graph."
            )
        if np.any(weights < 0):
            raise ValueError("All elements in 'weights' must be non-negative.")
    elif weights is not None:
        raise ValueError("'weights' must be a string or an array.")

    if weights is not None:
        edges = list(g.edges)
        edge_weight_dict = {edges[i]: weights[i] for i in range(len(edges))}
        nx.set_edge_attributes(g, edge_weight_dict, "weight")
        weight = "weight"
    else:
        weight = None

    if seed is not None:
        np.random.seed(seed)

    node_list = list(g.nodes)

    # Select random pivot nodes
    pivs = np.random.choice(node_list, pivots, replace=False)

    # Calculate the shortest path length from the pivots to all other nodes
    A = nx.to_scipy_sparse_array(g, weight=weight, nodelist=node_list, format="csr")

    # This is a workaround for what seems to be a bug in the type of
    # the indices of the sparse array created above
    A.indices = A.indices.astype(np.intc, copy=False)
    A.indptr = A.indptr.astype(np.intc, copy=False)
    D = sp.sparse.csgraph.shortest_path(
        A,
        directed=False,
        unweighted=weight is None,
        method="D",
        indices=np.where(np.isin(g.nodes, pivs))[0],
    ).T

    # Center values in rows and columns
    D2 = D**2
    cmean = np.mean(D2, axis=0)
    rmean = np.mean(D2, axis=1)
    D_pivs_centered = D2 - np.add.outer(rmean, cmean) + np.mean(D2)

    # Compute SVD and use distances to compute coordinates for all nodes
    # in an abstract cartesian space
    _, _, Vh = sp.sparse.linalg.svds(D_pivs_centered, k=dim, random_state=seed)

    coordinates = D_pivs_centered @ np.transpose(Vh)
    # Flip the coordinates here to make sure that we get the correct coordinate ordering
    # i.e. iqr(x) > iqr(y) > iqr(z)
    coordinates = np.flip(coordinates, axis=1)

    coordinates = {node_list[i]: coordinates[i, :] for i in range(coordinates.shape[0])}

    return coordinates


def _prob_edge_weights(
    g: nx.Graph,
    k: int = 5,
) -> np.ndarray:
    """Compute edge weights based on k-step transition probabilities.

    The transition probabilities are computed using powers of the
    stochastic matrix of the graph with self-loops allowed. Self-loops
    are necessary for bipartite graphs and without them many transitions
    will be impossible. For instance, if we have a bipartite graph with
    A nodes and B nodes where A can only be connected with B and vice versa,
    there is no possible 2-step path from an A node to a B node. However,
    if we allow self-loops, we can go from an A node to itself and then to
    a B node in two steps. By allowing self-loops, we make all transitions
    within the neighborhood k possible.

    The transition probabilities are computed for a k-step random walk
    by taking the k'th power of the stochastic matrix. Transition
    probabilities are not symmetric, i.e. it matters what node we start
    from. The probability of going from i to j in k steps
    is not the same as the probability of going from j to i in k steps.
    This is impractical for layout algorithms that require a single weight
    per edge. To make the transition probabilities symmetric, we multiply
    the transition probabilities in both directions (pk(i, j) * pk(j, i)).
    This way, we get the probability of going from i to j in k steps and then
    back again in k steps, so it no longer matters if we start in i or j.

    Once we have computed the transition probabilities for all k-step
    paths, we then extract the probabilities for the edges in the original
    graph.

    If we consider an edge (u, v) in the original graph, we now have
    the probability of a random walker going from u to v in k steps and then
    back again in k steps with self-loops allowed. If u anv v are well
    connected (if there are many possible k-step paths between them), this
    probability should be high.

    :param g: A networkx graph object
    :param k: The number of steps in the random walk
    :return: An array of edge weights
    :rtype: np.array
    :raises ValueError: if conditions are not met
    """
    # Check that k is an integer between 1 and 10
    if not isinstance(k, int) and k < 1 or k > 10:
        raise ValueError("'k' must be an integer between 1 and 10.")

    # Get the adjacency matrix. By default it is order by the nodes
    A = nx.to_scipy_sparse_array(g, weight=None, format="csr")

    # Add 1 to the diagonal to allow self-loops
    A = A + sp.sparse.diags_array([1] * A.shape[0], format="csr")

    # Divide by row sum to get the stochastic matrix
    D = sp.sparse.diags_array(1 / A.sum(axis=1), format="csr")
    P = D @ A

    # Compute the transition probabilities for a k-step walk
    min_weight = 0.001  # To avoid having the sparse matrix grow too dense
    P_step = _mat_pow(P, k, prune_threshold=min_weight)
    P_step = (P_step + min_weight * P) / (
        1 + min_weight
    )  # Ensure that the original values are not pruned

    # Keep edges from original graph
    P_step = P_step.multiply(A)

    # Compute bi-directional transition probabilities which are symmetric.
    # Now we get the probability of going from i to j and back again in k
    # steps, so it no longer matters if we start in i or j.
    P_step_bidirectional = P_step.multiply(P_step.T)

    # Extract the transition probabilities for edges in g
    edges = np.array(g.edges)
    nodes = np.array(g.nodes)

    # Get end node IDs for the edges
    node_from = edges[:, 0]
    node_to = edges[:, 1]

    # Find the correct positions to extract from the probability matrix
    index_dict = {val: idx for idx, val in enumerate(nodes)}
    vectorized_index = np.vectorize(lambda x: index_dict.get(x, -1))
    row_indices = vectorized_index(node_from)
    col_indices = vectorized_index(node_to)

    # Extract the probabilities
    edge_probs = np.asarray(P_step_bidirectional[row_indices, col_indices])

    return edge_probs


def _mat_pow(mat, power, prune_threshold: float | None = None):
    mat_power = mat.copy()
    for _ in range(power - 1):
        if prune_threshold:
            mat_power.data[np.abs(mat_power.data) < prune_threshold] = 0
            mat_power.eliminate_zeros()
        mat_power = mat @ mat_power
    return mat_power



"""Functions to create a summary interactive web report.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

from pixelator import __version__
from pixelator.pna.report.common import PixelatorPNAReporting
from pixelator.pna.report.models.report_metadata import SampleMetadata
from pixelator.pna.report.qcreport.builder import PNAQCReportBuilder
from pixelator.pna.report.qcreport.collect import collect_report_data
from pixelator.pna.report.qcreport.types import SampleInfo

logger = logging.getLogger("pixelator.report")


def create_qc_report(
    reporting: PixelatorPNAReporting,
    sample_name: str,
    info: SampleInfo,
    output_path: Path,
) -> None:
    """Create a dynamic customer QC report for a single sample.

    A helper function to create a dynamic web report of a single sample.
    The function uses a template stored in 'report/qcreport/template.html'
    as a base template which is filled with the different metrics and stats.

    :param reporting: A PixelatorReporting object
    :param sample_name: the name of the sample
    :param info: A dictionary containing meta information about the sample
    :param output_path: the path to the output folder
    """
    sample_id = info.sample_id
    logger.info("Creating QC report for sample %s", sample_id)

    report_data = collect_report_data(reporting, sample_id)

    # uses the default template
    builder = PNAQCReportBuilder()
    report_path = str(Path(output_path) / f"{sample_id}.qc-report.html")

    with open(report_path, "wb") as f:
        builder.write(f, sample_info=info, data=report_data)

    logger.debug("QC report created in %s", output_path)


def create_per_sample_qc_reports(
    input_path: Path,
    output_path: Path,
    panel: Optional[str],
    metadata: Optional[str],
    verbose: Optional[bool],
) -> None:
    """Parse stage metrics and build an interactive HTML report file.

    This function iterates the different steps (preqc, adapterqc, demux, collapse,
    graph and annotate) for a given processed dataset (folder). The function will parse
    the metrics of all the samples present in each subfolder (step). These metrics will
    then be converted to tables. The function will also generate interactive plotly
    figures using the output data present in the graph/annotate steps. All the tables
    and figures will be saved to files and used to generate individual html report for
    each sample (web reports).

    :param input_path: the path to results folder containing all the steps
    :param output_path: the path to the output folder
    :param panel: a path to a panel file or a panel name to load from the config
    :param metadata: a path to a metadata file
    :param verbose: run if verbose mode when true
    :returns: None
    :rtype: None
    :raises AssertionError: when the input is not valid
    :raises RuntimeError: when the metadata is not valid
    """
    from pixelator.pna.config import load_antibody_panel, pna_config

    logger.info("Generating PNA QC reports from %s", input_path)

    # collect and save metrics
    if panel is not None:
        panel_obj = load_antibody_panel(pna_config, panel)
    else:
        panel_obj = None

    reporting = PixelatorPNAReporting(Path(input_path))

    sample_metadata = SampleMetadata.from_csv(metadata) if metadata else None

    # create the dynamic report (one per sample)

    for sample in reporting.samples():
        this_sample_metadata = (
            sample_metadata.get_by_id(sample) if sample_metadata else None
        )
        sample_description = (
            this_sample_metadata.description if this_sample_metadata else ""
        )

        invocation_info = reporting.cli_invocation_info(sample)
        technology_version = ""

        try:
            collapse_design_option = invocation_info.get_option("collapse", "--design")
            technology_version = collapse_design_option.value
        except KeyError:
            logger.warning("No design info found for sample %s", sample)

        webreport_panel_name = None
        webreport_panel_version = None

        if panel_obj is not None:
            webreport_panel_name = (
                panel_obj.name if panel_obj.name else str(panel_obj.filename)
            )
            webreport_panel_version = panel_obj.version

        # create SampleInfo object
        sample_info = SampleInfo(
            pixelator_version=__version__,
            generation_date=datetime.now(timezone.utc).isoformat(),
            sample_id=sample,
            sample_description=sample_description,
            technology="PNA",
            technology_version=technology_version,
            panel_name=webreport_panel_name,
            panel_version=webreport_panel_version,
            parameters=list(invocation_info),  # type: ignore
        )

        create_qc_report(
            reporting,
            sample_name=sample,
            info=sample_info,
            output_path=output_path,
        )

    logger.info("Finished generating QC reports")



"""Model for report data returned by the single-cell layout stage.

Copyright © 2024 Pixelgen Technologies AB.
"""

from typing import Literal

from pixelator.pna.report.models.base import SampleReport


class LayoutSampleReport(SampleReport):
    """Model for report data returned by the single-cell layout stage."""

    report_type: Literal["layout"] = "layout"



"""Two-bit encoding of DNA sequences.

This module provides functions to encode/decode DNA sequences using two bits per nucleotide.

Copyright © 2024 Pixelgen Technologies AB.
"""

import numpy as np

_comp = str.maketrans("ACGT", "TGCA")

_STR2INT: dict[int, np.uint64] = {
    ord("A"): np.uint64(0),
    ord("C"): np.uint64(1),
    ord("G"): np.uint64(2),
    ord("T"): np.uint64(3),
}
_int2str: bytes = b"ACGT"


_STR2INT_4: dict[int, np.uint64] = {
    ord("A"): np.uint64(0b0110),
    ord("C"): np.uint64(0b0101),
    ord("T"): np.uint64(0b0011),
    ord("G"): np.uint64(0b0000),
    ord("N"): np.uint64(0b1111),
}
_int2str_4: bytes = b"ACTG"


def _int2chars(value: np.uint64) -> bytes:
    res = np.zeros(4, dtype=np.uint8)

    res[0] = _int2str[value & 3]
    value >>= np.uint64(8)
    res[1] = _int2str[value & 3]
    value >>= np.uint64(8)
    res[2] = _int2str[value & 3]
    value >>= np.uint64(8)
    res[3] = _int2str[value & 3]

    return res.tobytes()


def pack_2bits(kmer: bytes) -> np.uint64:
    """Pack a kmer into an int using two bits per nucleotide.

    A kmer must be at most 32 nucleotides long.
    Any nucleotides beyond the 32nd are ignored.

    :param kmer: the kmer to pack
    :return: the packed kmer as an integer
    """
    # pack the kmer into an int
    assert len(kmer) <= 32

    value = np.uint64(0)
    for c in kmer[32::-1]:
        value <<= np.uint(2)
        value |= _STR2INT[c]

    return np.uint64(value)


def pack_4bits(kmer: bytes) -> np.uint64:
    """Pack a kmer into an int using four bits per nucleotide.

    A kmer must be at most 16 nucleotides long.
    Any nucleotides beyond the 16th are ignored.

    :param kmer: the kmer to pack
    :return: the packed kmer as an integer
    """
    assert len(kmer) <= 16

    value = np.uint64(0)
    for c in kmer[:16]:
        value <<= np.uint(4)
        value |= _STR2INT_4[c]
    return np.uint64(value)


def unpack_2bits(packed: int, k: int) -> bytes:
    """Unpack a kmer from an integer using two bits per nucleotides.

    :param packed: the packed kmer as an integer
    :param k: the length of the kmer
    """
    np_packed = np.uint64(packed)
    seq = bytearray(k)
    for i in range(k):
        seq[i] = _int2str[np_packed & np.uint64(3)]
        np_packed >>= np.uint64(2)
    return bytes(seq)


def unpack_4bits(packed: int, k: np.uint64) -> bytes:
    """Unpack a kmer from an integer using two bits per nucleotides.

    :param packed: the packed kmer as an integer
    :param k: the length of the kmer
    """
    seq = bytearray(k)
    for i in range(k):
        seq[i] = _int2str_4[packed & 15]
        packed >>= 4
    return bytes(seq)



"""Console script for pixelator (collapse).

Copyright © 2022 Pixelgen Technologies AB.
"""

import sys
from collections import defaultdict
from concurrent import futures
from pathlib import Path

import click
import polars as pl

from pixelator.common.report.models import SummaryStatistics
from pixelator.common.utils import (
    create_output_stage_dir,
    get_process_pool_executor,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.cli.common import design_option, logger, output_option
from pixelator.mpx.collapse import collapse_fastq
from pixelator.mpx.config import config, get_position_in_parent, load_antibody_panel
from pixelator.mpx.report.models import CollapseSampleReport


def _handle_errors(jobs, executor):
    for job in jobs:
        exception = job.exception()
        if exception is None:
            continue

        logger.error(
            "Found an issue in the process pool. Trying to determine what went wrong and set the correct exit code. Exception was: %s",
            exception,
        )
        process_map = executor._processes
        for pid in process_map.keys():
            exit_code = process_map[pid].exitcode
            if exit_code is not None and exit_code != 0:
                logger.error(
                    "The child process in the process pool returned a non-zero exit code: %s.",
                    exit_code,
                )
                # If we have an out of memory exception, make sure we exit with that.
                if abs(exit_code) == 9:
                    logger.error(
                        "One of the child processes was killed (exit code: 9). "
                        "Usually this is caused by the out-of-memory killer terminating the process. "
                        "The parent process will return an exit code of 137 to indicate that it terminated because of a kill signal in the child process."
                    )
                    sys.exit(137)
        logger.error(
            "Was unable to determine what when wrong in process pool. Will raise original exception."
        )
        raise exception


@click.command(
    "collapse",
    short_help=(
        "collapse pixel data (FASTQ) by UMI-UPI to remove duplicates"
        " and perform error correction"
    ),
    options_metavar="<options>",
)
@click.argument(
    "input_files",
    nargs=-1,
    required=True,
    type=click.Path(exists=True),
    metavar="FASTQ_FILES",
)
@click.option(
    "--markers-ignore",
    required=False,
    default=None,
    type=click.STRING,
    help="A list of comma separated antibodies to ignore (discard)",
)
@click.option(
    "--algorithm",
    required=False,
    default="adjacency",
    type=click.Choice(["adjacency", "unique"]),
    show_default=True,
    help=(
        "The algorithm to use for collapsing (adjacency will perform error correction"
        " using the number of mismatches given)"
    ),
)
@click.option(
    "--max-neighbours",
    default=60,
    required=False,
    type=click.IntRange(1, 250),
    show_default=True,
    help=(
        "The maximum number of neighbors to use when searching for similar sequences."
        " This number depends on the sequence depth and the ratio of"
        " erronous molecules expected. A high value can make the algorithm slower."
        " This is only used when algorithm is set to 'adjacency'"
    ),
)
@click.option(
    "--mismatches",
    default=2,
    required=False,
    type=click.IntRange(0, 5),
    show_default=True,
    help=(
        "The number of mismatches allowed when collapsing. This is only used when the"
        " algorithm is set to 'adjacency'."
    ),
)
@click.option(
    "--min-count",
    default=1,
    required=False,
    type=click.IntRange(1, 10),
    show_default=True,
    help=(
        "Discard molecules with with a count (reads) lower than this (set to 1 to"
        " disable)"
    ),
)
@click.option(
    "--panel",
    required=True,
    type=str,
    help=(
        "A key of a panel file in the config,"
        " or a csv file with the antibody panel conjugations"
    ),
)
@output_option
@design_option
@click.pass_context
@timer
def collapse(
    ctx,
    input_files,
    markers_ignore,
    algorithm,
    max_neighbours,
    mismatches,
    min_count,
    panel,
    output,
    design,
):
    """Collapse Molecular Pixelation data (FASTQ) by umi-upi to remove duplicates and perform error correction."""  # noqa
    # log input parameters
    log_step_start(
        "collapse",
        output=output,
        markers_ignore=markers_ignore,
        algorithm=algorithm,
        max_neighbours=max_neighbours,
        mismatches=mismatches,
        min_count=min_count,
        panel=panel,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(
        input_files=input_files, allowed_extensions=("fastq.gz", "fq.gz")
    )

    # obtain a list of markers to ignore if given by the user
    if markers_ignore is not None:
        markers_ignore = markers_ignore.split(",")

    # update arguments from config file (CLI arguments have priority)
    try:
        assay = config.get_assay(design)

        if "umi-a" in assay.region_ids:
            umia_start, umia_end = get_position_in_parent(assay, "umi-a")
        else:
            umia_start, umia_end = (None, None)

        upia_start, upia_end = get_position_in_parent(assay, "upi-a")
        upib_start, upib_end = get_position_in_parent(assay, "upi-b")
        umib_start, umib_end = get_position_in_parent(assay, "umi-b")

    except KeyError as exc:
        raise click.ClickException(f"Parsing attribute from config file {exc}")

    # sanity check on upi and umi arguments
    if upia_end <= upia_start:
        click.ClickException("UPIA end or start positions seems to be incorrect")
    if upib_end <= upib_start:
        click.ClickException("UPIB end or start positions seems to be incorrect")
    if umia_start is None and umia_end is not None:
        click.ClickException("You must specify both the start and end position in UMIA")
    if umib_start is None and umib_end is not None:
        click.ClickException("You must specify both the start and end position in UMIB")
    if umia_start is None and umia_end is None:
        logger.info("UMIA will be ignored in the collapsing")
    elif umia_end <= umia_start:
        click.ClickException("UMIA end or start positions seems to be incorrect")
    if umib_start is None and umib_end is None:
        logger.info("UMIB will be ignored in the collapsing")
    elif umib_end <= umib_start:
        click.ClickException("UMIB end or start positions seems to be incorrect")

    # create output folder if it does not exist
    collapse_output = create_output_stage_dir(output, "collapse")

    # load marker panel
    panel = load_antibody_panel(config, panel)

    # group the input files by sample name
    input_samples = defaultdict(list)
    for fastq_file in input_files:
        sample = get_sample_name(fastq_file)
        input_samples[sample].append(fastq_file)

    if len(input_samples) == 0:
        click.ClickException("No input file could be found with a valid sample name")

    for sample, files in input_samples.items():
        logger.info(f"Processing {len(files)} files for sample {sample}")

        write_parameters_file(
            ctx,
            collapse_output / f"{sample}.meta.json",
            command_path="pixelator single-cell-mpx collapse",
        )

        # run cutadapt (demux mode) using parallel processing
        with get_process_pool_executor(
            nbr_cores=ctx.obj["CORES"], logging_setup=ctx.obj["LOGGER"]
        ) as executor:
            jobs = []
            for fastq_file in files:
                # get the marker from the file name
                sequence = fastq_file.split(".processed-")[1].split(".")[0]
                marker = panel.get_marker_id(sequence)
                if marker is None:
                    raise click.ClickException(
                        f"Retrieving marker from file {fastq_file}"
                    )

                # ignore the marker if it is in the ignore list
                if markers_ignore is not None and marker in markers_ignore:
                    logger.debug("Ignoring %s with antibody %s", fastq_file, marker)
                else:
                    logger.debug("Processing %s with antibody %s", fastq_file, marker)

                    jobs.append(
                        executor.submit(
                            collapse_fastq,
                            input_file=fastq_file,
                            algorithm=algorithm,
                            marker=marker,
                            sequence=sequence,
                            upia_start=upia_start,
                            upia_end=upia_end,
                            upib_start=upib_start,
                            upib_end=upib_end,
                            umia_start=umia_start,
                            umia_end=umia_end,
                            umib_start=umib_start,
                            umib_end=umib_end,
                            max_neighbours=max_neighbours,
                            mismatches=mismatches,
                            min_count=min_count,
                        )
                    )
            jobs = list(futures.as_completed(jobs))
            _handle_errors(jobs, executor)

        total_input_reads = 0
        tmp_files = []
        for job in jobs:
            # the worker returns a path to a file (temp antibody edge list)
            tmp_file, input_reads_count = job.result()
            if tmp_file is not None:
                tmp_files.append(tmp_file)
            if input_reads_count is not None:
                total_input_reads += input_reads_count

        # create the final edge list from all the temporary ones (antibodies)
        logger.debug("Creating edge list for sample %s", sample)
        df = pl.concat(
            (pl.read_ipc(f, memory_map=False) for f in tmp_files), how="vertical"
        )

        # Collect some stats for reporting
        collapsed_molecule_count_stats = SummaryStatistics.from_series(
            df["unique_molecules_count"]
        )

        output_file = collapse_output / f"{sample}.collapsed.parquet"
        # Remove the unique_molecules_count column before saving
        df.drop("unique_molecules_count")
        df.write_parquet(output_file, compression="zstd")

        # remove temporary edge list files
        for f in tmp_files:
            logger.debug("Removing temporary edge list %s", f)
            Path(f).unlink(missing_ok=True)

        output_read_count = int(df["count"].sum())
        molecule_count = int(df.shape[0])

        report = CollapseSampleReport(
            sample_id=sample,
            input_read_count=total_input_reads,
            output_read_count=output_read_count,
            molecule_count=molecule_count,
            collapsed_molecule_count_stats=collapsed_molecule_count_stats,
        )

        report.write_json_file(collapse_output / f"{sample}.report.json", indent=4)



"""Collect files and transform data for reporting from a :py:class:`PixelatorWorkdir`.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import json
import logging
import typing
from pathlib import Path

from pixelator.mpx.report.common.reporting import _ordered_pixelator_commands
from pixelator.pna.amplicon.report import AmpliconSampleReport
from pixelator.pna.analysis.report import AnalysisSampleReport
from pixelator.pna.collapse.independent import IndependentCollapseSampleReport
from pixelator.pna.collapse.report import CollapseSampleReport
from pixelator.pna.demux.report import DemuxSampleReport
from pixelator.pna.graph.report import GraphSampleReport
from pixelator.pna.layout.report import LayoutSampleReport
from pixelator.pna.report.common.cli_info import CLIInvocationInfo
from pixelator.pna.report.common.workdir import (
    PixelatorPNAWorkdir,
    SingleCellPNAStage,
    WorkdirOutputNotFound,
)
from pixelator.pna.report.models import (
    CommandInfo,
    ReadsDataflowReport,
)
from pixelator.pna.report.models.base import SampleReport

logger = logging.getLogger("pixelator.report")
ModelT = typing.TypeVar("ModelT", bound=type[SampleReport])


class PixelatorPNAReporting:
    """Collect files for reporting from a :py:class:`PixelatorWorkdir`.

    :ivar workdir: The pixelator output folder or a :py:class:`PixelatorWorkdir`
        instance
    """

    def __init__(self, workdir: Path | PixelatorPNAWorkdir):
        """Initialize the PixelatorReporting object.

        :param workdir: The pixelator output folder or a :py:class:`PixelatorWorkdir`
        """
        if isinstance(workdir, PixelatorPNAWorkdir):
            self.workdir = workdir
        else:
            self.workdir = PixelatorPNAWorkdir(workdir)

        # all pixelator commands in defined order
        self._command_list: list[str] | None = None

    def samples(self) -> set[str]:
        """Return a list of all samples encountered from command metadata files."""
        return self.workdir.samples()

    def amplicon_metrics(self, sample_name: str) -> AmpliconSampleReport:
        """Return the amplicon metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellPNAStage.AMPLICON, sample_name
        )
        return AmpliconSampleReport.from_json(sample_file)

    def demux_metrics(self, sample_name: str) -> DemuxSampleReport:
        """Return the demux metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellPNAStage.DEMUX, sample_name
        )
        return DemuxSampleReport.from_json(sample_file)

    def collapse_metrics(
        self, sample_name: str
    ) -> CollapseSampleReport | IndependentCollapseSampleReport:
        """Return the collapse metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellPNAStage.COLLAPSE, sample_name
        )
        with open(sample_file) as f:
            data = json.load(f)

            if data["report_type"] == "collapse-independent":
                return IndependentCollapseSampleReport.from_json(sample_file)
            else:
                return CollapseSampleReport.from_json(sample_file)

    def graph_metrics(self, sample_name: str) -> GraphSampleReport:
        """Return the graph metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellPNAStage.GRAPH, sample_name
        )
        return GraphSampleReport.from_json(sample_file)

    def analysis_metrics(self, sample_name: str) -> AnalysisSampleReport:
        """Return the analysis metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellPNAStage.ANALYSIS, sample_name
        )
        return AnalysisSampleReport.from_json(sample_file)

    def layout_metrics(self, sample_name: str) -> LayoutSampleReport:
        """Return the layout metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellPNAStage.LAYOUT, sample_name
        )
        return LayoutSampleReport.from_json(sample_file)

    def reads_flow(self, sample_name: str) -> ReadsDataflowReport:
        """Return a summary with the flow of read counts through the pipeline.

        :param sample_name: The sample to return the dataflow for
        :return ReadsDataflowReport: A class:`ReadsDataflowReport` instance
        """
        amplicon_metrics = self.amplicon_metrics(sample_name)
        demux_metrics = self.demux_metrics(sample_name)
        collapse_metrics = self.collapse_metrics(sample_name)
        graph_metrics = self.graph_metrics(sample_name)

        return ReadsDataflowReport(
            product_id="single-cell-pna",
            sample_id=sample_name,
            input_read_count=amplicon_metrics.input_reads,
            amplicon_output_read_count=amplicon_metrics.output_reads,
            demux_output_read_count=demux_metrics.output_reads,
            collapse_output_molecule_count=collapse_metrics.output_molecules,
            graph_output_molecule_count=graph_metrics.molecules_output,
        )

    def cli_invocation_info(self, sample: str, cache: bool = True) -> CLIInvocationInfo:
        """Return the commandline options used to invoke multiple pixelator commands on a sample.

        :param sample: The sample to return the commandline for
        :param cache: Use cached data if available
        :return: A `CLIInvocationInfo` to query commandline invocations.
        :raises WorkdirOutputNotFound: If no commandline metadata is found for the sample
        """
        # Function scope import to avoid circular dependencies

        try:
            metadata_files = self.workdir.metadata_files(sample, cache=cache)
        except WorkdirOutputNotFound as e:
            e.message = f'No command line metadata found for sample: "{sample}"'
            raise

        data_flat = []
        for f in metadata_files:
            command_info_flat = CommandInfo.from_json(f)
            data_flat.append(command_info_flat)

        if self._command_list is None:
            self._command_list = _ordered_pixelator_commands()

        # mypy cannot detect that self._command_list cannot be None
        # anymore at this point so we cast it explicitly.
        try:
            data_flat.sort(
                key=lambda x: typing.cast("list[str]", self._command_list).index(
                    x.command
                )
            )
        except ValueError as e:
            # Do not fail if the command is not in the list,
            # just log a warning and continue.
            unknown_commands = {x.command for x in data_flat} - set(self._command_list)
            logger.warning(
                f"Unknown command in meta.json file: {','.join(unknown_commands)}"
            )

        return CLIInvocationInfo(data_flat, sample_id=sample)  # type: ignore[arg-type]



"""
Console script for pixelator (common functions)

Copyright © 2022 Pixelgen Technologies AB.
"""

import collections
import functools
import logging.handlers
from pathlib import Path
from typing import Dict, Mapping, Optional

import click

BASE_DIR = str(Path(__file__).parent)
logger = logging.getLogger("pixelator.cli")


# code snippet obtained from
# https://stackoverflow.com/questions/47972638/how-can-i-define-the-order-of-click-sub-commands-in-help
# the purpose is to order subcommands in order of addition
class OrderedGroup(click.Group):
    """Custom click.Group that keeps insertion order for subcommands."""

    def __init__(  # noqa: D107
        self,
        name: Optional[str] = None,
        commands: Optional[Dict[str, click.Command]] = None,
        **kwargs,
    ):
        super(OrderedGroup, self).__init__(name, commands, **kwargs)
        self.commands = commands or collections.OrderedDict()

    def list_commands(  # type: ignore
        self, ctx: click.Context
    ) -> Mapping[str, click.Command]:
        """Return a list of subcommands."""
        return self.commands


# snippet adapted from
# https://stackoverflow.com/questions/46765803/python-click-multiple-group-names
class AliasedOrderedGroup(OrderedGroup):
    """Custom click.Group that supports aliases.

    Currently only supports aliases for subgroups.
    """

    def group(self, *args, **kwargs):
        """Attach a click group that supports aliases."""

        def decorator(f):
            aliased_group = []
            aliases = kwargs.pop("aliases", [])
            main_group = super(AliasedOrderedGroup, self).group(*args, **kwargs)(f)

            for alias in aliases:
                grp_kwargs = kwargs.copy()
                del grp_kwargs["name"]
                grp = super(AliasedOrderedGroup, self).group(
                    alias, *args[1:], **grp_kwargs
                )(f)
                grp.short_help = "Alias for '{}'".format(main_group.name)
                aliased_group.append(grp)

            # for all the aliased groups, link to all attributes from the main group
            for aliased in aliased_group:
                aliased.commands = main_group.commands
                aliased.params = main_group.params
                aliased.callback = main_group.callback
                aliased.epilog = main_group.epilog
                aliased.options_metavar = main_group.options_metavar
                aliased.add_help_option = main_group.add_help_option
                aliased.no_args_is_help = main_group.no_args_is_help
                aliased.hidden = main_group.hidden
                aliased.deprecated = main_group.deprecated

            return main_group

        return decorator


def output_option(func):
    """Wrap a Click entrypoint to add the --output option."""

    @click.option(
        "--output",
        required=True,
        type=click.Path(exists=False),
        help=(
            "The path where the results will be placed (it is created if it does not"
            " exist)"
        ),
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def design_option(func):
    """Decorate a click command and add the --design option."""
    from pixelator.mpx.config import config

    assay_options = list(config.assays.keys())

    @click.option(
        "--design",
        required=True,
        default=None,
        type=click.Choice(assay_options),
        help="The design to load from the configuration file",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper



"""Helper typehints for the pixelator package.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import os
from pathlib import Path, PurePath
from typing import Union

# type alias for path-like objects
PathType = Union[str, Path, PurePath, os.PathLike]



"""Pipeline console script for denoising pna data.

Copyright © 2025 Pixelgen Technologies AB.
"""

import logging
from pathlib import Path

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.pna import read
from pixelator.pna.analysis.denoise import DenoiseOneCore
from pixelator.pna.analysis.report import DenoiseSampleReport
from pixelator.pna.analysis_engine import AnalysisManager, LoggingSetup
from pixelator.pna.cli.common import output_option
from pixelator.pna.pixeldataset.io import PxlFile

logger = logging.getLogger(__name__)


@click.command(
    "denoise",
    short_help=("Denoise a pxl file by applying node filtering techniques."),
    options_metavar="<options>",
)
@click.argument(
    "pxl_file",
    required=True,
    type=click.Path(exists=True),
    metavar="<PIXELFILE>",
)
@click.option(
    "--run-one-core-graph-denoising",
    required=False,
    is_flag=True,
    help="Run the denoise step to remove markers that are over-expressed in the one-core layer of a component.",
)
@click.option(
    "--one-core-ratio-threshold",
    default=0.9,
    required=False,
    type=click.FloatRange(
        0,
        1,
    ),
    show_default=True,
    help=(
        "ratio of the number of nodes in the one-core layer to the total number of nodes in a component. "
        "If the ratio is above this threshold, the component is marked as disqualified for denoising."
    ),
)
@click.option(
    "--pval-threshold",
    default=0.05,
    required=False,
    type=click.FloatRange(
        0,
        1,
    ),
    show_default=True,
    help="pvalue threshold for an over-expression in the one-core layer to be considered significant.",
)
@click.option(
    "--inflate-factor",
    default=1.5,
    required=False,
    type=click.FloatRange(
        1,
        10,
    ),
    show_default=True,
    help="How much to inflate number of noise markers in the one-core layer to remove.",
)
@output_option
@click.pass_context
@timer
def denoise(
    ctx,
    pxl_file,
    run_one_core_graph_denoising,
    one_core_ratio_threshold,
    pval_threshold,
    inflate_factor,
    output,
):
    """Denoise components of a PXL file."""
    input_files = [pxl_file]
    log_step_start(
        "denoise",
        input_files=input_files,
        run_one_core_graph_denoising=run_one_core_graph_denoising,
        one_core_ratio_threshold=one_core_ratio_threshold,
        pval_threshold=pval_threshold,
        inflate_factor=inflate_factor,
        output=output,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files=input_files, allowed_extensions=("pxl",))

    sample_name = get_sample_name(pxl_file)
    pxl_file = PxlFile(Path(pxl_file))
    denoise_output = create_output_stage_dir(output, "denoise")
    output_file = denoise_output / f"{sample_name}.denoised_graph.pxl"
    output_pxl_file_target = PxlFile.copy_pxl_file(pxl_file, output_file)

    write_parameters_file(
        ctx,
        denoise_output / f"{sample_name}.meta.json",
        command_path="pixelator single-cell-pna denoise",
    )
    metrics = denoise_output / f"{sample_name}.report.json"

    if not run_one_core_graph_denoising:
        report = DenoiseSampleReport(
            sample_id=sample_name,
            product_id="single-cell-pna",
            number_of_umis_removed=None,
            ratio_of_umis_removed=None,
            number_of_disqualified_components=None,
            ratio_of_disqualified_components=None,
        )
        report.write_json_file(metrics, indent=4)
        return

    analysis_to_run = [
        DenoiseOneCore(pval_threshold, inflate_factor, one_core_ratio_threshold)
    ]
    logging_setup = LoggingSetup.from_logger(ctx.obj.get("LOGGER"))
    manager = AnalysisManager(analysis_to_run, logging_setup=logging_setup)
    pxl_dataset = read(pxl_file.path)

    pxl_dataset_denoised = manager.execute(pxl_dataset, output_pxl_file_target)

    number_of_umis_removed = int(
        pxl_dataset_denoised.adata().obs["number_of_nodes_removed_in_denoise"].sum()
    )
    ratio_of_umis_removed = float(
        number_of_umis_removed
        / (pxl_dataset.adata().obs["n_umi"].sum() + number_of_umis_removed)
    )
    number_of_disqualified_components = int(
        pxl_dataset_denoised.adata().obs["disqualified_for_denoising"].sum()
    )
    ratio_of_disqualified_components = float(
        pxl_dataset_denoised.adata().obs["disqualified_for_denoising"].mean()
    )
    report = DenoiseSampleReport(
        sample_id=sample_name,
        product_id="single-cell-pna",
        number_of_umis_removed=number_of_umis_removed,
        ratio_of_umis_removed=ratio_of_umis_removed,
        number_of_disqualified_components=number_of_disqualified_components,
        ratio_of_disqualified_components=ratio_of_disqualified_components,
    )

    report.write_json_file(metrics, indent=4)



"""Implements a CombiningModifier interface for a Pipeline.

A modifier must be callable and typically implemented as a class with a
__call__ method.

Copyright © 2024 Pixelgen Technologies AB.
"""

from abc import ABC, abstractmethod

from cutadapt.info import ModificationInfo
from dnaio import SequenceRecord


class CombiningModifier(ABC):
    """Interface for a modifier that combines two reads into a single read."""

    @abstractmethod
    def __call__(
        self,
        read1: SequenceRecord,
        read2: SequenceRecord,
        info1: ModificationInfo,
        info2: ModificationInfo,
    ) -> SequenceRecord | None:
        """Combine two reads into a single read."""
        pass



"""Model for report data returned by the single-cell collapse stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import pydantic

from pixelator.common.report.models.summary_statistics import SummaryStatistics
from pixelator.pna.report.models.base import SampleReport


class CollapseSampleReport(SampleReport):
    """Model for data returned by the demux stage."""

    input_read_count: int = pydantic.Field(
        ...,
        description="The total number of input reads before collapsing.",
    )

    output_read_count: int = pydantic.Field(
        ...,
        description="The total number of reads from unique molecules after collapsing.",
    )

    molecule_count: int = pydantic.Field(
        ...,
        description="The total number of unique molecules in the graph.",
    )

    collapsed_molecule_count_stats: SummaryStatistics = pydantic.Field(
        ...,
        description="Summary statistics for the number of unique molecules that are collapsed into a single molecule.",
    )

    @pydantic.computed_field(  # type: ignore
        return_type=float,
        description="The fraction of reads (after preprocessing) that are PCR duplicates.",
    )
    @property
    def fraction_duplicate_reads(self) -> float:  # noqa: D102
        return 1.0 - (self.molecule_count / self.output_read_count)



"""Marker panel management for different Molecular Pixelation assays.

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

import warnings
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

import pandas as pd
import pydantic
import ruamel.yaml as yaml

from pixelator.common.types import PathType
from pixelator.common.utils import logger

if TYPE_CHECKING:
    from pixelator.common.config import Config


class AntibodyPanelMetadata(pydantic.BaseModel):
    """Class representing the metadata of a Molecular Pixelation antibody panel."""

    model_config = pydantic.ConfigDict(extra="ignore")

    version: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    aliases: List[str] = []


class AntibodyPanel:
    """Class representing a Molecular Pixelation antibody panel."""

    # required columns
    _REQUIRED_COLUMNS = [
        "marker_id",
        "control",
        "nuclear",
        "sequence",
        "conj_id",
    ]

    # and these should have unique values
    _UNIQUE_COLUMNS = [
        "marker_id",
        "sequence",
    ]

    def __init__(
        self,
        df: pd.DataFrame,
        metadata: AntibodyPanelMetadata,
        file_name: Optional[str] = None,
    ) -> None:
        """Load a panel from a dataframe and metadata.

        :param df: The dataframe containing the panel information.
        :param metadata: The metadata for the panel.
        :param file_name: The optional basename of the file from which
            the panel is loaded.

        :returns: None
        :raises AssertionError: exception if panel file is missing,
                                invalid or with incorrect format
        """
        self._filename = file_name
        self._metadata = metadata
        self._df = df

        # validate the panel
        errors = self.validate_antibody_panel(df)
        if len(errors) > 0:
            msg_str = "\n".join(errors)
            raise AssertionError(
                f"The following errors were found validating the panel: {msg_str}"
            )

    @classmethod
    def from_csv(cls, filename: PathType) -> "AntibodyPanel":
        """Create an AntibodyPanel from a csv panel file.

        :param filename: The path to the panel file.
        :returns: The AntibodyPanel object.
        :raises AssertionError: exception if panel file is missing,
        :rtype: AntibodyPanel
        """
        panel_file = Path(filename)

        if not panel_file.is_file() or panel_file.suffix != ".csv":
            raise AssertionError(
                f"Panel file {filename} not found or has an incorrect format"
            )

        logger.debug("Creating Antibody panel from file %s", filename)

        df = cls._parse_panel(panel_file)
        metadata = cls._parse_header(panel_file)

        logger.debug("Antibody panel from file %s created", filename)

        return cls(df, metadata, file_name=panel_file.name)

    @property
    def name(self) -> Optional[str]:
        """The name defined in the panel metadata."""
        return self._metadata.name

    @property
    def version(self) -> Optional[str]:
        """Return the panel file version."""
        return self._metadata.version

    @property
    def description(self) -> Optional[str]:
        """Return the panel file description."""
        return self._metadata.description

    @property
    def aliases(self) -> list[str]:
        """Return the (optional) list of panel file aliases."""
        return self._metadata.aliases

    @classmethod
    def validate_antibody_panel(self, panel_df: pd.DataFrame) -> list[str]:
        """Perform validation on an antibody panel file.

        Will try to find as many issues as possible.

        This will not directly raise the issue (since that makes it difficult
        to find multiple problems at once) instead it will return a list of str
        (one for each issue).

        Usage example:
        ```
        >>> errors = panel.validate_antibody_panel(panel_df)
        ... if len(errors) > 0:
        ...     AssertionError("There was a problem with the panel data!")
        ````

        :param panel_df: DataFrame with data of the panel to validate
        :returns list[str]: a list of errors (str)
        """
        errors = []

        # some basic sanity check on the panel size and columns
        if not set(self._REQUIRED_COLUMNS).issubset(set(panel_df.columns)):
            missing_columns = set(self._REQUIRED_COLUMNS) - set(panel_df.columns)
            errors.append(f"Panel has missing required columns: {missing_columns}")
            return errors

        if panel_df.shape[0] == 0:
            errors.append("Panel file is empty")
            return errors

        # sanity check on the unique columns
        for col in self._UNIQUE_COLUMNS:
            if not len(panel_df[col].unique()) == len(panel_df[col]):
                errors.append(f"All values in column: {col} were not unique")

        return errors

    @classmethod
    def _parse_panel(cls, panel_file: Path) -> pd.DataFrame:
        panel = pd.read_csv(str(panel_file), comment="#")

        # validate the panel
        errors = cls.validate_antibody_panel(panel)
        if len(errors) > 0:
            msg_str = "\n".join(errors)
            raise AssertionError(
                f"The following errors were found validating the panel: {msg_str}"
            )

        panel = cls._transform_legacy_panels(panel)

        # assign the sequence (unique) as index
        panel.index = panel.sequence  # type: ignore

        # return a local copy
        return panel.copy()

    @classmethod
    def _transform_legacy_panels(cls, df: pd.DataFrame) -> pd.DataFrame:
        """Transform legacy panels to the new format.

        :param df: DataFrame with data of the panel to validate
        :returns pd.DataFrame: The in-place modified input dataframe
        """
        # update control and nuclear column to boolean
        TR_TABLE = {"(?i)yes": "True", "(?i)no": "False"}

        df["control"] = (
            df["control"]
            .astype("string[pyarrow]")
            .fillna("")
            .replace(TR_TABLE, regex=True)
            .astype(bool)
        )
        df["nuclear"] = (
            df["nuclear"]
            .astype("string[pyarrow]")
            .fillna("")
            .replace(TR_TABLE, regex=True)
            .astype(bool)
        )

        return df

    @classmethod
    def _parse_header(cls, file: Path) -> AntibodyPanelMetadata:
        """Parse front-matter YAML from a file.

        :param file: the file to parse
        :return AntibodyPanelMetadata: a pydantic model with the metadata
        """
        yaml_loader = yaml.YAML(typ="safe")

        metadata_lines = []
        with open(str(file), "r") as f:
            for line in f:
                if line.startswith("# "):
                    metadata_lines.append(line[2:])
                else:
                    break

        metadata = "".join(metadata_lines)
        raw_config = list(yaml_loader.load_all(metadata))

        if len(raw_config) == 0:
            warnings.warn(f"Expected a YAML frontmatter in {file}")
            return AntibodyPanelMetadata(version=None, name=None, description=None)

        frontmatter = raw_config[0]
        return AntibodyPanelMetadata.model_validate(frontmatter)

    @cached_property
    def markers_control(self) -> List[str]:
        """Return a list of marker control (names)."""
        return list(self._df[self._df["control"]].marker_id.unique())

    @cached_property
    def markers(self) -> List[str]:
        """Return the list of unique markers in the panel."""
        return list(self._df.marker_id.unique())

    @property
    def df(self) -> pd.DataFrame:
        """Return the panel dataframe."""
        return self._df

    @property
    def filename(self) -> Optional[str]:
        """Return the filename of the marker panel."""
        return self._filename

    @cached_property
    def size(self) -> int:
        """Return the size of the marker panel."""
        return self._df.shape[0]

    def get_marker_id(self, seq: str) -> str:
        """Return the marker name."""
        return self._df.loc[seq].marker_id


def load_antibody_panel(config: Config, panel: PathType) -> AntibodyPanel:
    """Load an antibody panel from a file or from the config file.

    :param config: the config object
    :param panel: the path to the panel file or the name of the
        panel in the config file

    :returns: the antibody panel
    :rtype: AntibodyPanel
    """
    panel_str = str(panel)
    panel_from_config = config.get_panel(panel_str)

    if panel_from_config is not None:
        logger.info("Found panel in config file: %s", panel_from_config.name)
        return panel_from_config

    panel_obj = AntibodyPanel.from_csv(panel)
    return panel_obj



"""Module contains various useful graph functions.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
import typing
import warnings
from typing import Dict, List, Literal, Optional, Tuple, Union

import networkx as nx
import pandas as pd
import polars as pl
import xxhash
from scipy.sparse.linalg import matrix_power

from pixelator.common.graph import Graph
from pixelator.common.graph.backends.implementations import (
    NetworkXGraphBackend,
)
from pixelator.common.report.models import SummaryStatistics

logger = logging.getLogger(__name__)


def union(graphs: List[Graph]) -> Graph:
    """Create union of graphs.

    Create a union of the provided graphs, merging any vertices
    which share the same name.

    :param graphs: the graphs to create the union from
    :return: a new graph that is the union of the input `graphs`
    :rtype: Graph
    :raises: AssertionError if not all underlying graphs have the same backend type.
    """
    backends = [type(g._backend) for g in graphs]
    if not all(map(lambda b: backends[0] == b, backends)):
        raise AssertionError("All graph objects must share the same backend")

    if backends[0] == NetworkXGraphBackend:
        return Graph(
            backend=NetworkXGraphBackend.from_raw(
                nx.union_all(
                    [graph._backend.raw for graph in graphs],
                    rename=[f"g{idx}-" for idx, _ in enumerate(graphs)],
                )
            )
        )

    raise NotImplementedError()


def components_metrics(edgelist: pd.DataFrame) -> pd.DataFrame:
    """Calculate metrics per component.

    A helper function that computes a dataframe of metrics for
    each component in the data present in the edge list given
    as input (component column). The metrics include: vertices,
    edges, markers, upis, degree mean and max.
    :param edgelist: an edge list dataframe with a membership column
    :returns: a pd.DataFrame with the metrics per component
    :rtype: pd.DataFrame
    :raises: AssertionError when the input edge list is not valid
    """
    if "component" not in edgelist.columns:
        raise AssertionError("Edge list is missing the membership column")

    logger.debug(
        "Computing components metrics for edge list with %i edges", edgelist.shape[0]
    )

    cmetrics = []
    index = []
    # iterate the components to obtain the metrics of each component
    for component_id, group_df in edgelist.groupby("component", observed=True):
        # compute metrics
        a_pixels = group_df["upia"].nunique()
        b_pixels = group_df["upib"].nunique()
        pixels = a_pixels + b_pixels
        antibodies = group_df["marker"].nunique()
        molecules = group_df.shape[0]

        reads = group_df["count"].sum()
        mean_reads_per_molecule = group_df["count"].mean()
        median_reads_per_molecule = group_df["count"].median()

        # Please note that we need to use observed=True
        # here upia is a categorical column, and since not
        # all values are present in all components, this is
        # required to get a correct value.
        b_pixels_per_a_pixel_series = group_df.groupby("upia", observed=True)[
            "upib"
        ].nunique()
        mean_b_pixels_per_a_pixel = b_pixels_per_a_pixel_series.mean()
        median_b_pixels_per_a_pixel = b_pixels_per_a_pixel_series.median()

        a_pixels_per_b_pixel = group_df.groupby("upib", observed=True)["upia"].nunique()
        mean_a_pixels_per_b_pixel = a_pixels_per_b_pixel.mean()
        median_a_pixels_per_b_pixel = a_pixels_per_b_pixel.median()

        # Same reasoning as above
        molecule_count_per_a_pixel_series = group_df.groupby("upia", observed=True)[
            "umi"
        ].count()
        mean_molecules_per_a_pixel = molecule_count_per_a_pixel_series.mean()
        median_molecules_per_a_pixel = molecule_count_per_a_pixel_series.median()

        a_pixel_b_pixel_ratio = a_pixels / b_pixels

        cmetrics.append(
            (
                pixels,
                a_pixels,
                b_pixels,
                antibodies,
                molecules,
                reads,
                mean_reads_per_molecule,
                median_reads_per_molecule,
                mean_b_pixels_per_a_pixel,
                median_b_pixels_per_a_pixel,
                mean_a_pixels_per_b_pixel,
                median_a_pixels_per_b_pixel,
                a_pixel_b_pixel_ratio,
                mean_molecules_per_a_pixel,
                median_molecules_per_a_pixel,
            )
        )
        index.append(component_id)

    # create components metrics data frame
    components_metrics = pd.DataFrame(
        index=pd.Index(index, name="component"),
        columns=[
            "pixels",
            "a_pixels",
            "b_pixels",
            "antibodies",
            "molecules",
            "reads",
            "mean_reads_per_molecule",
            "median_reads_per_molecule",
            "mean_b_pixels_per_a_pixel",
            "median_b_pixels_per_a_pixel",
            "mean_a_pixels_per_b_pixel",
            "median_a_pixels_per_b_pixel",
            "a_pixel_b_pixel_ratio",
            "mean_molecules_per_a_pixel",
            "median_molecules_per_a_pixel",
        ],
        data=cmetrics,
    )

    logger.debug("Component metrics computed")
    return components_metrics


def _get_extended_adjacency(graph: Graph, k: int = 0):
    A = graph.get_adjacency_sparse()
    A.setdiag(1, k=0)
    An = (matrix_power(A, k) > 0).astype(int)
    return An


def _get_neighborhood_counts(
    node_marker_counts,
    graph,
    k: int = 0,
    normalization: Optional[Literal["mean"]] = None,
):
    An = _get_extended_adjacency(graph, k=k)
    neighbourhood_counts = An @ node_marker_counts

    # TODO Optionally add more methods here
    if normalization == "mean":
        nbr_of_neighbors_per_node = An.sum(axis=1)
        # Reshape to ensure broadcasting compatibility
        nbr_of_neighbors_per_node = nbr_of_neighbors_per_node.reshape((-1, 1))
        neighbourhood_counts = neighbourhood_counts / nbr_of_neighbors_per_node

    df = pd.DataFrame(
        data=neighbourhood_counts,
        columns=node_marker_counts.columns.copy(),
        index=node_marker_counts.index.copy(),
    )
    df.columns.name = "markers"
    df.index.name = "node"

    return df


def create_node_markers_counts(
    graph: Graph,
    k: int = 0,
    normalization: Optional[Literal["mean"]] = None,
) -> pd.DataFrame:
    """Create a matrix of marker counts for each in the graph.

    A helper function that computes and returns a data frame of antibody counts per
    node (vertex) of the graph given as input (preferably a fully connected component).
    The parameter k allows to include neighbors (of each node) when computing the
    counts (using `agg_func` to aggregate the counts). K defines the number of levels
    when searching neighbors. The graph must contain a vertex attribute called 'markers'
    which is dictionary of marker counts per vertex.

    :param graph: a graph (preferably a connected component)
    :param k: number of neighbors to include per node (0 no neighbors,
              1 first level, ...)
    :param normalization: selects a normalization method to apply when
                          building neighborhoods

    :returns: a pd.DataFrame with the antibody counts per node
    :rtype: pd.DataFrame
    """
    if k == 0 and normalization:
        warnings.warn(
            (
                f"Using `normalization={normalization}` when k=0 "
                "has no effect, since no neighborhood is created."
            )
        )

    node_marker_counts = graph.node_marker_counts

    if k == 0:
        return node_marker_counts

    neighborhood_counts = _get_neighborhood_counts(
        node_marker_counts=node_marker_counts,
        graph=graph,
        k=k,
        normalization=normalization,
    )
    return neighborhood_counts


class EdgelistMetrics(typing.TypedDict, total=True):
    """TypedDict for edgelist metrics."""

    component_count: int
    molecule_count: int
    marker_count: int
    read_count: int
    a_pixel_count: int
    b_pixel_count: int

    read_count_per_molecule_stats: SummaryStatistics

    fraction_molecules_in_largest_component: float
    fraction_pixels_in_largest_component: float

    edges_with_colliding_upi_count: int
    edges_removed_in_multiplet_recovery_first_iteration: int
    edges_removed_in_multiplet_recovery_refinement: int
    fraction_edges_removed_in_refinement: float


MetricsDict = typing.TypeVar(
    "MetricsDict", Dict[str, Union[int, float]], EdgelistMetrics
)


def _edgelist_metrics_lazyframe(edgelist: pl.LazyFrame) -> EdgelistMetrics:
    metrics: EdgelistMetrics = {}  # type: ignore

    unique_counts = edgelist.select(
        pl.col("upia").n_unique(),
        pl.col("upib").n_unique(),
        pl.col("marker").n_unique(),
    ).collect()

    metrics["a_pixel_count"] = int(unique_counts["upia"][0])
    metrics["b_pixel_count"] = int(unique_counts["upib"][0])
    metrics["marker_count"] = int(unique_counts["marker"][0])
    # Note that we get upi here and count that, because otherwise just calling count
    # here confuses polars since there is a column with that name.
    metrics["molecule_count"] = int(
        edgelist.select(pl.col("upia").count()).collect()["upia"][0]
    )

    counts_per_molecule = edgelist.select(pl.col("count")).collect()["count"]
    metrics["read_count"] = int(counts_per_molecule.sum())
    metrics["read_count_per_molecule_stats"] = SummaryStatistics.from_series(
        counts_per_molecule
    )

    component_stats = (
        edgelist.group_by("component")
        .agg(
            pl.col("upia").n_unique().alias("n_upia"),
            pl.col("upib").n_unique().alias("n_upib"),
            pl.len().alias("n_molecule"),
        )
        .collect()
    )
    metrics["component_count"] = component_stats.shape[0]
    component_stats = component_stats.with_columns(
        n_upi=pl.col("n_upia") + pl.col("n_upib")
    )
    largest_component = component_stats.filter(
        pl.col("n_molecule") == component_stats["n_molecule"].max()
    )[0]
    metrics["fraction_molecules_in_largest_component"] = (
        largest_component["n_molecule"][0] / component_stats["n_molecule"].sum()
    )
    metrics["fraction_pixels_in_largest_component"] = (
        largest_component["n_upi"][0] / component_stats["n_upi"].sum()
    )
    return metrics


def edgelist_metrics(
    edgelist: pl.DataFrame | pd.DataFrame | pl.LazyFrame,
) -> EdgelistMetrics:
    """Compute edgelist metrics.

    A simple function that computes a dictionary of basic metrics
    from an edge list (pl.DataFrame).

    :param edgelist: the edge list (pl.DataFrame)
    :param graph: optionally add the graph instance that corresponds to the
                  edgelist (to not have to re-compute it)
    :returns: a dataclass of metrics
    :rtype: EdgelistMetrics
    :raises TypeError: if edgelist is not either a pl.LazyFrame
    , pl.DataFrame or a pd.DataFrame
    """
    if isinstance(edgelist, pl.LazyFrame):
        logger.debug("Computing edgelist metrics where edgelist type is pl.LazyFrame")
        return _edgelist_metrics_lazyframe(edgelist)

    if isinstance(edgelist, pd.DataFrame):
        logger.debug("Computing edgelist metrics where edgelist type is pd.DataFrame")
        edgelist = pl.from_pandas(edgelist)

    if isinstance(edgelist, pl.DataFrame):
        metrics: EdgelistMetrics = {}  # type: ignore

        metrics["a_pixel_count"] = edgelist.n_unique("upia")
        metrics["b_pixel_count"] = edgelist.n_unique("upib")
        metrics["marker_count"] = edgelist.n_unique("marker")
        metrics["molecule_count"] = edgelist.shape[0]

        counts_per_molecule = edgelist["count"]
        metrics["read_count"] = int(counts_per_molecule.sum())
        metrics["read_count_per_molecule_stats"] = SummaryStatistics.from_series(
            counts_per_molecule
        )

        component_stats = edgelist.group_by("component").agg(
            pl.col("upia").n_unique().alias("n_upia"),
            pl.col("upib").n_unique().alias("n_upib"),
            pl.len().alias("n_molecule"),
        )
        metrics["component_count"] = component_stats.shape[0]
        component_stats = component_stats.with_columns(
            n_upi=pl.col("n_upia") + pl.col("n_upib")
        )
        largest_component = component_stats.filter(
            pl.col("n_molecule") == component_stats["n_molecule"].max()
        )[0]
        metrics["fraction_molecules_in_largest_component"] = (
            largest_component["n_molecule"][0] / component_stats["n_molecule"].sum()
        )
        metrics["fraction_pixels_in_largest_component"] = (
            largest_component["n_upi"][0] / component_stats["n_upi"].sum()
        )
        return metrics
    raise TypeError(
        "edgelist was not of type `pd.LazyFrame`, `pd.DataFrame` or `pl.DataFrame"
    )


def map_upis_to_components(
    edgelist: pl.LazyFrame,
    node_component_map: pd.Series,
    node_depth_map: Optional[pd.Series] = None,
) -> pl.LazyFrame:
    """Update the edgelist with component names corresponding to upia/upib.

    Using the node_component_map, this function will add the component
    information to the edgelist. Two columns are added to the edgelist,
    i.e. component_a and component_b respectively for upia and upib.
    The component names are then determined by calculating a hash of
    the nodes in that component.

    :param edgelist: the edge list
    :param node_component_map: a pd.Series mapping the nodes to their components
    :returns: the remaining_edgelist and the removed_edgelist
    :rtype: pl.LazyFrame
    :raises TypeError: if edgelist is not either a pl.LazyFrame or a pd.DataFrame
    """
    # Create a mapping of the components to a hash of its UPIs
    node_component_map = node_component_map.astype(str)
    components = node_component_map.groupby(node_component_map)
    for _, comp in components:
        comp_nodes = sorted(comp.index)
        comp_hash = xxhash.xxh64()
        for node in comp_nodes:
            comp_hash.update(str(node))
        node_component_map[comp_nodes] = comp_hash.hexdigest()
    node_component_dict = node_component_map.to_dict()
    logger.debug("Mapping components on the edge list")
    edgelist_with_component_info = edgelist.with_columns(
        component_a=pl.col("upia")
        .cast(pl.String)
        .replace_strict(
            node_component_dict,
            default="",
        ),
    ).with_columns(
        component_b=pl.col("upib")
        .cast(pl.String)
        .replace_strict(
            node_component_dict,
            default="",
        ),
    )
    if node_depth_map is not None:
        node_depth_dict = node_depth_map.to_dict()
        edgelist_with_component_info = (
            edgelist_with_component_info.with_columns(
                upia_depth=pl.col("upia")
                .cast(pl.String)
                .replace_strict(node_depth_dict, default=0),
                upib_depth=pl.col("upib")
                .cast(pl.String)
                .replace_strict(node_depth_dict, default=0),
            )
            .with_columns(
                depth=pl.min_horizontal([pl.col("upia_depth"), pl.col("upib_depth")])
            )
            .drop(["upia_depth", "upib_depth"])
        )

    return edgelist_with_component_info


def update_edgelist_membership(
    edgelist: pl.LazyFrame | pd.DataFrame,
    node_component_map: Optional[pd.Series] = None,
) -> pl.LazyFrame | pd.DataFrame:
    """Update the edgelist with component names.

    Using the node_component_map, this function will add the component
    information to the edgelist. If for an edge, components of UPIA and
    UPIB do not match, that edge will be removed. If node_component_map
    is missing, it will be constructed based on the connected component
    in the graph made from the edgelist.

    :param edgelist: the edge list
    :param node_component_map: a pd.Series mapping the nodes to their components
    if missing, it will be constructed based on the connected components in the
    graph made from the edgelist.
    :returns: the remaining_edgelist and the removed_edgelist
    :rtype: pl.LazyFrame | pd.DataFrame
    :raises TypeError: if edgelist is not either a pl.LazyFrame or a pd.DataFrame
    """
    if isinstance(edgelist, pd.DataFrame):
        was_dataframe = True
        edgelist = pl.LazyFrame(edgelist)
    else:
        was_dataframe = False

    if node_component_map is None:
        edges = (
            edgelist.select(["upia", "upib"])
            .group_by(["upia", "upib"])
            .len()
            .sort(["upia", "upib"])  # sort to make sure the graph is the same
            .collect()
        )
        graph = nx.from_edgelist(edges.select(["upia", "upib"]).iter_rows())

        node_component_map = pd.Series(index=graph.nodes())
        for i, cc in enumerate(nx.connected_components(graph)):
            node_component_map[list(cc)] = i
        del graph

    if isinstance(edgelist, pl.LazyFrame):
        logger.debug("Updating edgelist where type is pl.LazyFrame")
        if "component" in edgelist.collect_schema().names():
            logger.info("The input edge list already contains a component column")

        edgelist_with_component_info = map_upis_to_components(
            edgelist, node_component_map
        )
        edgelist_with_component_info, _ = split_remaining_and_removed_edgelist(
            edgelist_with_component_info
        )
    else:
        raise TypeError("edgelist was not of type pl.LazyFrame or pd.DataFrame")

    if was_dataframe:
        return edgelist_with_component_info.collect().to_pandas()
    else:
        return edgelist_with_component_info


def split_remaining_and_removed_edgelist(
    edgelist: pl.LazyFrame,
) -> Tuple[pl.LazyFrame, pl.LazyFrame]:
    """Split the edgelist into remaining and removed.

    Inputs an edgelist with component_a and component_b columns.
    If the two columns match in a row and they are not invalid,
    i.e. empty string. They will be added to the remaining_edgelist.
    Otherwise they will be added to the removed_edgelist.

    :param edgelist: the edge list
    :returns: the remaining_edgelist and the removed_edgelist
    """
    if "component" in edgelist.collect_schema().names():
        logger.info("The input edge list already contains a component column")
        edgelist = edgelist.drop("component")

    remaining_edgelist = (
        edgelist.filter(pl.col("component_a") != "")
        .filter(pl.col("component_a") == pl.col("component_b"))
        .rename({"component_a": "component"})
        .drop("component_b")
    )
    columns = remaining_edgelist.collect_schema().names()
    if "depth" in columns:
        remaining_edgelist = remaining_edgelist.drop("depth")

    removed_edgelist = edgelist.filter(
        (pl.col("component_a") == "") | (pl.col("component_a") != pl.col("component_b"))
    )
    return remaining_edgelist, removed_edgelist



"""Integration tests for the annotate command.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
from pathlib import Path

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseAnnotateTestsMixin(BaseWorkflowTestMixin):
    """Base class for annotate command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_annotate_run"])
    ```
    """

    __stage_key__ = "annotate"

    @pytest.mark.dependency(name="test_annotate_run", depends=["test_graph_run"])
    def test_annotate_run(self):
        """Run method for the test."""
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")
        panel = self.__get_data("panel")
        panel_file = self.__get_data("panel_file")

        input_files = list(
            map(lambda f: str(f), (self.workdir / "graph").glob("*.edgelist.parquet"))
        )

        command = [
            "--log-file",
            "annotate-pixelator.log",
            "single-cell-mpx",
            "annotate",
            "--panel",
            str(Path(panel_file).name) if panel_file else panel,
            "--output",
            ".",
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        command += input_files

        file_deps = list(input_files)
        if panel_file is not None:
            file_deps.append(panel_file)

        self.context.run_command("annotate", command, file_deps)
        with open(str(self.workdir / "annotate-pixelator.log")) as f:
            self.context.set_logs("annotate", f.read())

        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_annotate_run"])
    def test_annotate_dataset_exists(self):
        """Check that the pixel file is created."""
        dataset_files = (self.workdir / "annotate").glob("*.annotate.dataset.pxl")
        for f in dataset_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_annotate_run"])
    def test_annotate_report_exists(self):
        """Check that the json report exists."""
        json_files = (self.workdir / "annotate").glob("*.report.json")
        for f in json_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_annotate_run"])
    def test_annotate_images_exist(self):
        """Check that the rank size plot has been created."""
        png_files = (self.workdir / "annotate").glob("*.rank_vs_size.png")

        if "--verbose" not in self.__this_command:
            return pytest.skip(
                'Skipping rank_vs_size image test, ( "--verbose" not set )'
            )

        for f in png_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_annotate_run"])
    def test_raw_component_metrics_exist(self):
        """Check that the raw metrics exists."""
        metric_files = (self.workdir / "annotate").glob(
            "*.raw_components_metrics.csv.gz"
        )

        if "--verbose" not in self.__this_command:
            return pytest.skip(
                'Skipping raw_components_metrics.csv.gz test, ( "--verbose" not set )'
            )

        for f in metric_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_annotate_run"])
    def test_doublet_calling_vertex_communities_exists(self):
        """Check that the vertex community file has been created."""
        vertex_files = (self.workdir / "annotate").glob("*.vertex_communities.csv.gz")

        if "--doublet-calling" not in self.__this_command:
            return pytest.skip(
                "Skipping vertex_communities.csv.gz test, "
                '( "--doublet-calling" not set )'
            )

        if "--verbose" not in self.__this_command:
            return pytest.skip(
                'Skipping raw_components_metrics.csv.gz test, ( "--verbose" not set )'
            )

        for f in vertex_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_annotate_run"])
    def test_umap_png_exists(self):
        """Check that the umap png exists."""
        png_files = (self.workdir / "annotate").glob("*.umap.png")

        if "--verbose" not in self.__this_command:
            return pytest.skip('Skipping umap.png test, ( "--verbose" not set )')

        if "Skipping clustering" in self.__this_logs:
            return pytest.skip(
                'Skipping umap.png test, ( "Skipping clustering" in output )'
            )

        for f in png_files:
            assert f.is_file()



"""Copyright © 2023 Pixelgen Technologies AB."""

from pixelator.common.config.panel import AntibodyPanelMetadata
from pixelator.common.config.utils import load_yaml_file
from pixelator.pna.config.assay import (
    AssayModel,
    PNAAssay,
    PNARegionType,
    Region,
    RegionModel,
    SequenceType,
    get_position_in_parent,
)
from pixelator.pna.config.config_class import Config, load_assays_package
from pixelator.pna.config.config_instance import pna_config
from pixelator.pna.config.panel import (
    PNAAntibodyPanel,
    load_antibody_panel,
)

__all__ = [
    "AssayModel",
    "RegionModel",
    "PNARegionType",
    "SequenceType",
    "PNAAssay",
    "Region",
    "get_position_in_parent",
    "Config",
    "pna_config",
    "load_assays_package",
    "load_yaml_file",
    "PNAAntibodyPanel",
    "AntibodyPanelMetadata",
    "load_antibody_panel",
]



"""Collect statistics about the collapse process.

Copyright © 2024 Pixelgen Technologies AB.
"""

import dataclasses
import typing
from pathlib import Path, PurePath
from typing import Any

import polars as pl
import pydantic

from pixelator.common.report.models import SummaryStatistics


@dataclasses.dataclass(slots=True)
class CollapseInputFile:
    """Keep track of the input file to collapse.

    :param path: Path to the input file.
    :param file_size: The total size of the input file.
    :param molecule_count: The number of rows in the dataframe.
    """

    path: str
    file_size: int
    molecule_count: int


class MarkerLinkGroupStats(pydantic.BaseModel):
    """Collect statistics of groups of molecules that will be collapsed.

    Attributes
    ----------
    corrected_reads_count: int
        The total number of reads in unique molecules that were error corrected to another "close" molecule.
    cluster_size_distribution: 1D array of ints
        The result of binning the size of each group of close molecules.
    collapsed_molecules_count: int
        The total number of unique molecules (UMI1+UMI2+UEI) after error correction.

    """

    marker_1: str
    marker_2: str

    input_reads_count: int = 0
    input_molecules_count: int = 0

    corrected_reads_count: int = 0
    cluster_size_distribution: list = pydantic.Field(
        ...,
        default_factory=list,
        description="The result of binning the size of each group of close molecules.",
    )
    collapsed_molecules_count: int = 0
    unique_marker_links_count: int = 0

    read_count_per_collapsed_molecule_stats: SummaryStatistics | None = None
    read_count_per_unique_marker_link_stats: SummaryStatistics | None = None
    uei_count_per_unique_marker_link_stats: SummaryStatistics | None = None


class CollapseSummaryStatistics(pydantic.BaseModel):
    """Collect summary statistics on reads and ueis."""

    read_counts_stats: SummaryStatistics | None = None
    uei_stats: SummaryStatistics | None = None

    @staticmethod
    def from_lazy_frame(collapsed_lz_df: pl.LazyFrame):
        """Create a CollapseSummaryStatistics from a collapsed LazyFrame."""
        df = collapsed_lz_df.select("uei_count", "read_count").collect()
        uei_stats = SummaryStatistics.from_series(df.get_column("uei_count"))
        read_stats = SummaryStatistics.from_series(df.get_column("read_count"))

        return CollapseSummaryStatistics(
            read_counts_stats=read_stats,
            uei_stats=uei_stats,
        )


class CollapseStatistics:
    """Collect statistics about the collapse process."""

    def __init__(self) -> None:
        """Initialize the statistics collector."""
        self._processed_files: list[CollapseInputFile] = []
        self._marker_pair_data: dict[tuple[str, str], dict[str, Any]] = {}
        self._summary_statistics: CollapseSummaryStatistics | None = None

    @typing.overload
    def add_input_file(
        self, input_file: PurePath, molecule_count: int, file_size: int
    ) -> None: ...

    @typing.overload
    def add_input_file(
        self, input_file: Path, molecule_count: int, file_size: int | None = None
    ) -> None: ...

    def add_input_file(
        self,
        input_file: Path | PurePath,
        molecule_count: int,
        file_size: int | None = None,
    ) -> None:
        """Collect file statistics for an input file to the MoleculeCollapser.

        :param input_file: The input file to collapse.
        :param molecule_count: The number of molecules in the input file
        :raise TypeError: If file_size is not provided when input_file is a PurePath.
        """
        if file_size is None and isinstance(input_file, Path):
            file_size = input_file.stat(follow_symlinks=True).st_size
        elif file_size is None:
            raise TypeError("file_size must be provided if input_file is a PurePath")

        self._processed_files.append(
            CollapseInputFile(
                path=str(input_file),
                file_size=file_size,
                molecule_count=molecule_count,
            )
        )

    def add_marker_stats(
        self,
        marker1: str,
        marker2: str,
        *,
        cluster_stats: MarkerLinkGroupStats,
        elapsed_time: float | None = None,
    ) -> None:
        """Add statistics for a marker pair.

        :param marker1: The first marker in the pair.
        :param marker2: The second marker in the pair.
        :param input_molecules_count: The number of unique molecules (UMI1+UMI2+UEI) before error correction.
        :param input_reads_count: The number of reads in unique molecules before error correction.
        :param cluster_stats: The statistics for the marker pair.
        :param elapsed_time: The time taken to process the marker pair.
        """
        key = (marker1, marker2)
        if key in self._marker_pair_data:
            raise KeyError("Marker pair already exists")

        self._marker_pair_data[key] = {
            **cluster_stats.model_dump(),
        }

        if elapsed_time is not None:
            self._marker_pair_data[key]["elapsed_real_time"] = elapsed_time

    def add_summary_statistics(self, collapsed_lz_df: pl.LazyFrame) -> None:
        """Add summary statistics for the entire collapse process.

        :param collapsed_lz_df: The collapsed dataframe.
        """
        self._summary_statistics = CollapseSummaryStatistics.from_lazy_frame(
            collapsed_lz_df
        )

    def get_high_level_statistics(self) -> dict:
        """Compute high level summary statistics from all the marker pairs."""
        reads_input = sum(
            data["input_reads_count"] for data in self._marker_pair_data.values()
        )
        molecules_input = sum(
            data["input_molecules_count"] for data in self._marker_pair_data.values()
        )
        molecules_output = sum(
            data["collapsed_molecules_count"]
            for data in self._marker_pair_data.values()
        )
        corrected_reads = sum(
            data["corrected_reads_count"] for data in self._marker_pair_data.values()
        )
        unique_marker_links_count = sum(
            data["unique_marker_links_count"]
            for data in self._marker_pair_data.values()
        )

        stats = {
            "input_reads": reads_input,
            "input_molecules": molecules_input,
            "output_molecules": molecules_output,
            "corrected_reads": corrected_reads,
            "unique_marker_links": unique_marker_links_count,
        }

        if self._summary_statistics:
            stats["summary_statistics"] = self._summary_statistics.dict()

        return stats

    def to_dict(self) -> dict:
        """Return the statistics as a dictionary."""
        flat_markers_data = list(self._marker_pair_data.values())
        return {
            **self.get_high_level_statistics(),
            "processed_files": [dataclasses.asdict(f) for f in self._processed_files],
            "markers": flat_markers_data,
        }



"""Copyright © 2023 Pixelgen Technologies AB."""

import logging
from pathlib import Path

import pytest

from pixelator.common.config.panel import load_antibody_panel
from pixelator.mpx.config import config

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseDemuxTestsMixin(BaseWorkflowTestMixin):
    """Base class for demux command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_demux_run"])
    ```
    """

    __stage_key__ = "demux"

    @pytest.mark.dependency(name="test_demux_run", depends=["test_adapterqc_run"])
    def test_demux_run(self):
        design = self.__get_data("design")
        panel = self.__get_data("panel")
        panel_file = self.__get_data("panel_file")
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")
        input_files = list(
            map(
                lambda f: str(f),
                (self.workdir / "adapterqc").glob("*.processed.fastq.gz"),
            )
        )

        command = [
            "--log-file",
            "demux-pixelator.log",
            "single-cell-mpx",
            "demux",
            "--panel",
            str(Path(panel_file).name) if panel_file else panel,
            "--output",
            ".",
            "--design",
            design,
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        command += input_files

        file_deps = list(input_files)
        if panel_file is not None:
            file_deps.append(panel_file)

        self.context.run_command("demux", command, file_deps)
        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_demux_run"])
    def test_demux_logfile_exist(self):
        assert (self.workdir / "adapterqc-pixelator.log").is_file()

    @pytest.mark.dependency(depends=["test_demux_run"])
    def test_demux_results_folder_exists(self):
        assert (self.workdir / "preqc").is_dir()

    @pytest.mark.dependency(depends=["test_demux_run"])
    def test_demux_processed_output_exists(self):
        panel_key_or_file = self.__get_data("panel") or self.__get_options(
            "common"
        ).get("panel_file")
        panel = load_antibody_panel(config, panel_key_or_file)

        for marker in panel.markers:
            processed_files = (self.workdir / "demux").glob(
                f"*.processed-{marker}.fastq.gz"
            )
            for f in processed_files:
                assert f.is_file()

    @pytest.mark.dependency(depends=["test_demux_run"])
    def test_demux_failed_output_exists(self):
        failed_files = (self.workdir / "demux").glob("*.failed.fastq.gz")
        for f in failed_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_demux_run"])
    def test_demux_json_output_exists(self):
        json_files = (self.workdir / "demux").glob("*.report.json")
        for f in json_files:
            assert f.is_file()
